from .constants.modules.olga import load_model as load_model
from .constants.modules.olga import generation_probability as pgen
from .constants.modules.olga import sequence_generation as seq_gen
from .constants.preprocessing import format_chain
from os import path
from multiprocessing import Pool, cpu_count

import pandas as pd

chainmap = {'A':'alpha','B':'beta'}
DIR = '/'.join(path.dirname(path.abspath(__file__)).split('/')[:-1]) + '/'

def load_models(chain='B'):

    chain = format_chain(chain)
            
    params_file_name = path.join(DIR,f'snetcr/constants/modules/olga/default_models/human_T_{chainmap[chain]}/model_params.txt')
    marginals_file_name = path.join(DIR,f'snetcr/constants/modules/olga/default_models/human_T_{chainmap[chain]}/model_marginals.txt')
    V_anchor_pos_file = path.join(DIR,f'snetcr/constants/modules/olga/default_models/human_T_{chainmap[chain]}/V_gene_CDR3_anchors.csv')
    J_anchor_pos_file = path.join(DIR,f'snetcr/constants/modules/olga/default_models/human_T_{chainmap[chain]}/J_gene_CDR3_anchors.csv')
    
    if chain == 'B':
        genomic_data = load_model.GenomicDataVDJ()
        generative_model = load_model.GenerativeModelVDJ()
    elif chain == 'A':
        genomic_data = load_model.GenomicDataVJ()
        generative_model = load_model.GenerativeModelVJ()

    genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
    generative_model.load_and_process_igor_model(marginals_file_name)

    return generative_model, genomic_data

def init_pgen_model(chain='B'):
    """
    Initialize the OLGA model for PGEN calculations.
    """
    generative_model, genomic_data = load_models(chain=chain)
    return pgen.GenerationProbabilityVDJ(generative_model, genomic_data)

def init_seqgen_model(chain='B'):
    generative_model, genomic_data = load_models(chain=chain)
    if chain == 'B':
        return seq_gen.SequenceGenerationVDJ(generative_model, genomic_data)
    else:
        return seq_gen.SequenceGenerationVJ(generative_model, genomic_data)

def compute_pgen_for_sequence(pgen_model, seq):
    """
    Compute the generation probability for a single sequence.
    """
    cdr3, v, j = seq
    return pgen_model.compute_aa_CDR3_pgen(cdr3, v, j)

def calc_pgen(sequences, chain='B', ncpus=1):
    """
    Calculate the average generation probability of a cluster.
    PGEN calculations are based on the OLGA module.
    """
    
    if ncpus > cpu_count():
        print('Number of CPUs requested is greater than available CPUs. Using all available CPUs.')
        ncpus = cpu_count()

    pgen_model = init_pgen_model(chain=chain)

    with Pool(ncpus) as pool:
        results = pool.starmap(compute_pgen_for_sequence, [(pgen_model, seq) for seq in sequences])
    
    return results

def generate_sequence(seq_gen_model):
    return seq_gen_model.gen_rnd_prod_CDR3()

def generate_sequences(n, chain='B'):
    """
    Generate random sequences based on the OLGA model.
    """
    
    seq_gen_model = init_seqgen_model(chain=chain)
    
    vmap = vgenes(chain=chain)
    jmap = jgenes(chain=chain)
    
    seqs = pd.DataFrame(
        [generate_sequence(seq_gen_model) for seq in range(n)]
        )
    seqs.columns = ['junction','junction_aa','v_call','j_call']
    seqs['v_call'] = seqs['v_call'].map(vmap)
    seqs['j_call'] = seqs['j_call'].map(jmap)
    seqs['locus'] = chain
    return seqs

def vgenes(chain='B'):
    '''
    Load V gene reference file that contains mapping information to extract V alleles from indices.
    '''
    _, genomic_data = load_models(chain=chain)
    return {i:v[0] for i,v in enumerate(genomic_data.genV)}

def jgenes(chain='B'):
    _, genomic_data = load_models(chain=chain)
    return {i:j[0] for i,j in enumerate(genomic_data.genJ)}



#from basic import *
from . import logo_tools
from .genetic_code import genetic_code


bases_plus = list(logo_tools.nucleotide_classes_lower_case.keys())

for a in bases_plus:
    for b in bases_plus:
        for c in bases_plus:
            codon = a+b+c
            if codon in genetic_code: continue

            aas = []
            for a1 in logo_tools.nucleotide_classes_lower_case[a]:
                for b1 in logo_tools.nucleotide_classes_lower_case[b]:
                    for c1 in logo_tools.nucleotide_classes_lower_case[c]:
                        aas.append( genetic_code[ a1+b1+c1 ] )
            if min(aas) == max(aas):
                genetic_code[codon] = aas[0]
            else:
                genetic_code[codon] = 'X'



def get_translation( seq, frame='+1' ): ## STUPID-- frame is 1/2/3 with a +/- in front
    assert frame[0] in '+-'
    if frame[0] == '-': seq = logo_tools.reverse_complement( seq )
    offset = abs( int( frame ))-1
    assert offset in range(3)
    seq = seq[offset:].lower() # note we take lowercase for compat w/code
    naa = len(seq)//3
    protseq = ''
    for i in range(naa):
        codon = seq[3*i:3*i+3]
        if '#' in codon:
            protseq += '#'
        else:
            protseq += genetic_code.get( codon, 'X' )
    return protseq

def get_translation_and_codons( seq, frame='+1' ): ## STUPID-- frame is 1/2/3 with a +/- in front
    assert frame[0] in '+-'
    if frame[0] == '-': seq = logo_tools.reverse_complement( seq )
    offset = abs( int( frame ))-1
    assert offset in range(3)
    seq = seq[offset:].lower()
    naa = len(seq)//3
    protseq = ''
    codons = []
    for i in range(naa):
        codon = seq[3*i:3*i+3]
        codons.append( codon )
        if '#' in codon:
            protseq += '#'
        else:
            protseq += genetic_code.get( codon, 'X' )
    return protseq,codons




import numpy as np
import faiss

from timeit import default_timer as timer
from scipy.sparse import coo_matrix
from .encoding import TCRDistEncoder
from .indexing import convert_range_search_output

def compute_sparse_distance_matrix(tcrs, chain, organism, exact=True, r=96.5, m=8):
    '''
    Compute the sparse distance matrix for a set of TCRs.
    '''
    # Encode the TCRs
    t0 = timer()
    start = timer()
    encoder = TCRDistEncoder(aa_dim=m,organism=organism,chain=chain).fit()
    vecs = encoder.transform(tcrs).astype(np.float32)
    print(f'Encoding TCRs took {timer()-start:.2f}s')
    d = vecs.shape[1]

    if exact:
        # Flat index will ensure 'exact' search
        # However, this is still an approximation of the true TCRdist
        start = timer()
        index = faiss.IndexFlatL2(d)
        index.add(vecs)
        print(f'Building the index took {timer()-start:.2f}s')
    else:

        if len(vecs) > 100000:
            n = int(len(vecs)/10)
        else:
            n = min(len(vecs), 10000) 

        nlist = int(np.sqrt(len(vecs))) # number of clusters
        quantizer = faiss.IndexFlatL2(d) # the quantizer is a flat (L2) index
        index = faiss.IndexIVFFlat(quantizer, d, nlist, faiss.METRIC_L2)
        index.train(vecs[np.random.choice(len(vecs), size=n, replace=False)])
        index.add(vecs)

    # Search index
    start = timer()
    lims, D, I = index.range_search(vecs, thresh=r)
    print(f'Searching index within radius {d} took {timer()-start:.2f}s')

    # Format the results.
    start = timer() 
    res = convert_range_search_output(lims, D, I)

    # Determine the size of the matrix
    max_index = len(tcrs)
    base = [([i,i],0) for i in range(max_index)]
    res += base
    res = list(set((tuple(indices), distance-1) for indices, distance in res))

    # Initialize lists to hold row indices, column indices, and data values
    rows, cols, values = [], [], []

    # Populate the lists with the given distances
    for indices, distance in res:
        i, j = indices
        rows.append(i)
        cols.append(j)
        values.append(distance)
        if i != j:  # Ensure the matrix is symmetric
            rows.append(j)
            cols.append(i)
            values.append(distance)

    # Convert the lists to a sparse matrix using coo_matrix
    dm = coo_matrix((values, (rows, cols)), shape=(max_index, max_index), dtype=np.float64)
    print(f'Building the sparse matrix took {timer()-start:.2f}s')
    print(f'Total time to compute distance matrix: {timer()-t0:.2f}s')

    return dm

# LEGACY
# def compute_sparse_distance_matrix(tcrs, chain, organism, exact=True, d=96.5, m=8):
#     '''
#     Compute the sparse distance matrix for a set of TCRs.
#     '''
#     # Encode the TCRs
#     t0 = timer()
#     start = timer()
#     encoder = TCRDistEncoder(aa_dim=m,organism=organism,chain=chain).fit()
#     vecs = encoder.transform(tcrs).astype(np.float32)
#     print(f'Encoding TCRs took {timer()-start:.2f}s')

#     if exact:
#         # Flat index will ensure 'exact' search
#         # However, this is still an approximation of the true TCRdist
#         start = timer()
#         index = FlatIndex(encoder=encoder)
#         index.add(tcrs, vecs)
#         print(f'Building the index took {timer()-start:.2f}s')
#     else:
#         # IVF index
#         start = timer()
#         k = round(len(vecs)/1000)
#         n = round(k/20)+2
#         if n > k:
#             n = k
#         index = IvfIndex(encoder=encoder, n_centroids=k, n_probe=n)
#         index.add(tcrs, vecs)
#         print(f'Building the index took {timer()-start:.2f}s')

#     # Search index
#     start = timer()
#     lims, D, I = index.idx.range_search(vecs, thresh=d)
#     print(f'Searching index within radius {d} took {timer()-start:.2f}s')

#     # Format the results.
#     start = timer() 
#     res = convert_range_search_output(lims, D, I)

#     # Determine the size of the matrix
#     max_index = len(tcrs)
#     base = [([i,i],0) for i in range(max_index)]
#     res += base
#     res = list(set((tuple(indices), distance-1) for indices, distance in res))

#     # Initialize lists to hold row indices, column indices, and data values
#     rows, cols, values = [], [], []

#     # Populate the lists with the given distances
#     for indices, distance in res:
#         i, j = indices
#         rows.append(i)
#         cols.append(j)
#         values.append(distance)
#         if i != j:  # Ensure the matrix is symmetric
#             rows.append(j)
#             cols.append(i)
#             values.append(distance)

#     # Convert the lists to a sparse matrix using coo_matrix
#     dm = coo_matrix((values, (rows, cols)), shape=(max_index, max_index), dtype=np.float64)
#     print(f'Building the sparse matrix took {timer()-start:.2f}s')
#     print(f'Total time to compute distance matrix: {timer()-t0:.2f}s')

#     return dm


import matplotlib.pyplot as plt
import logomaker
import pandas as pd
import numpy as np
import networkx as nx

from .neighbors import compute_sparse_distance_matrix

def cdr3_logo(cluster, ax, n_trim=0, c_trim=0):

    if c_trim > 0:
        seqs = cluster.junction_aa.apply(lambda x: x[n_trim:-c_trim])
    else:
        seqs = cluster.junction_aa.apply(lambda x: x[n_trim:])
    seqlen = len(seqs.iloc[0])
    aas = []
    for i in seqs:
        aas += i
    aas = sorted(list(set(aas)))

    pos_matrix = np.zeros((seqlen,len(aas)))
    aa_dict = {i:[0]*seqlen for i in aas}

    for pos in range(seqlen):
        aacount = seqs.str[pos].value_counts()
        for i in aacount.index:
            aa_dict[i][pos] = aacount[i]
    pos_df = pd.DataFrame(aa_dict)

    # create Logo object
    logo = logomaker.Logo(
        pos_df,
        ax=ax,
        font_name='Arial',
        color_scheme='dmslogo_funcgroup'
        )

    return pos_df

def plot_cluster(cluster, r, chain='B', organism='human', ax=None):

    matrix = compute_sparse_distance_matrix(cluster, chain=chain, organism=organism, d=r+.5)

    non_zero_values = matrix.data
    row_indices, col_indices = matrix.nonzero()

    mask = (non_zero_values >= 0)
    filtered_row_indices = row_indices[mask]
    filtered_col_indices = col_indices[mask]
    ids = np.column_stack((filtered_row_indices, filtered_col_indices))

    nodes = cluster.reset_index(drop=True).index.values
    edges = [(nodes[i[0]], nodes[i[1]]) for i in ids] 

    G = nx.Graph()
    G.add_nodes_from(list(nodes))
    G.add_edges_from(edges)
    pos = nx.spring_layout(G)

    coordinates = np.array(list(pos.values()))
    x = coordinates[:, 0]
    y = coordinates[:, 1]
    c = -np.log10(cluster.evalue)

    if ax is None:
        nx.draw_networkx_edges(G, pos, alpha=0.5)
        cbar = plt.scatter(x,y,linewidths=0.5,edgecolor='black',alpha=1,c=c,cmap='viridis')
    else:
        nx.draw_networkx_edges(G, pos, ax=ax, alpha=0.5)
        cbar = ax.scatter(x,y,linewidths=0.5,edgecolor='black',alpha=1,c=c,cmap='viridis')

    plt.colorbar(cbar, ax=ax)


import numpy as np
import matplotlib.pyplot as plt

def normal(x, mu, sigma):
    """
    Mathematical formulation of the normal distribution.

    Parameters
    ----------
    x
        x values of the distribution.
    mu
        Mean of the distribution.
    sigma
        Standard deviation around mu.

    """
    return 1/(sigma*np.sqrt(2*np.pi))*np.exp(-(x-mu)**2/(2*sigma**2))

def exponential(x, beta):
    """
    Mathematical formulation of an exponential function.

    Parameters
    ----------
    x
        x values of the distribution.
    beta
        Scale parameter.
    """
    return (1/beta)*(np.exp(-x/beta))


def gaussian_vector(nbins=64, mu=1, sigma=.1, n=10000):
    """
    Create a vector of ordered, binned probability densities from the normal distribution.

    Parameters
    ----------
    nbins
        Number of data points.
    mu
        Mean of the distribution.
    sigma
        Standard deviation around mu.
    n
        Number of samples to draw from the normal distribution.
    """
    s = np.random.normal(mu, sigma, n)
    count, bins, ignored = plt.hist(s, nbins-1, density=True)
    plt.close()
    return np.array(normal(x=bins, mu=mu, sigma=sigma))

def negexp_vector(nbins, beta=1, n=10000):
    """
    Create a vector of ordered, binned probability densities from a parametrized negative exponential distribution.

    Parameters
    ----------
    nbins
        Number of data points.
    beta
        Scale parameter.
    n
        Number of samples to draw from the exponential distribution.
    """
    s = np.random.exponential(beta, n)
    count, bins, ignored = plt.hist(s, nbins-1, density=True)
    plt.close()
    return np.array(exponential(bins, beta)) 


import pandas as pd
import numpy as np

from .constants.preprocessing import IMGT, adaptive_to_imgt_human, adaptive_to_imgt_mouse
from .constants.parsing import (
    _is_cdr3,
    format_junction,
    )

class Repertoire():
    '''
    Repertoire formatter.
    '''
    
    def __init__(self,data,organism='human'):
        self.data = data
        self.organism = organism

    def filter_and_format_single(
            self,
            cdr3aa_col = 'junction_aa',
            cdr3nt_col = 'junction',
            vgene_col = 'v_call',
            jgene_col = 'j_call',
            remove_nonfunctional=True
            ):
        
        '''
        Filter and format the data. Filtering includes:
        - Removing ambiguous CDR3 amino acid sequences
        - Removing ORF and non-functional genes
        - Parsing junction (CDR3 nucleotide sequence)
        '''

        col_remap = {
            cdr3nt_col:"junction",
            cdr3aa_col:"junction_aa",
            vgene_col:"v_call",
            jgene_col:"j_call",
            }
        
        self.data = self.data.rename(columns=col_remap)
            
        if self.data.v_call.str.contains('TCRBV').any():
            print('Detected Adaptive format')
            self.data = self.data[self.data.frame_type == 'In']
            self.data["v_imgt"] = self.data.v_call.map(adaptive_to_imgt_human)
            self.data["j_call"] = self.data.j_call.map(adaptive_to_imgt_human)
            # Correct the V allele if available else use IMGT mapping
            self.data = self.data.dropna(subset=["v_imgt","junction_aa","junction","j_call"])
            self.data["v_call"] = [self.data.v_imgt.iloc[n].split("*")[0]+f"*0{int(i)}" if not np.isnan(i) else self.data.v_imgt.iloc[n] for n,i in enumerate(self.data.v_allele)]
            self.data = self.data.drop(columns=["v_imgt"])

        # Remove ambiguous CDR3 amino acid sequences
        print("Remove ambiguous CDR3 amino acid sequences")
        self.data = self.data[self.data.junction_aa.apply(lambda cdr3: _is_cdr3(cdr3))]
        
        # Remove ORF and non-functional genes
        print("Parsing V/J genes")
        if remove_nonfunctional:
            functional = IMGT[IMGT['fct'].isin(['F','(F)','[F]'])]
            self.data = self.data[self.data.v_call.isin(functional.imgt_allele_name)]
        else:
            # make sure all V genes have IMGT annotation
            self.data = self.data[self.data.v_call.isin(IMGT.imgt_allele_name)]
        
        # Parse junction (CDR3 nucleotide sequence)
        print("Parsing CDR3 nucleotide sequence")
        self.data = format_junction(self.data)
        self.data = self.data.dropna(subset=["junction_aa","junction","v_call","j_call"])

        return self.data
    
    def filter_and_format_paired(
            self,
            cdr3aa_a_col = 'cdr3a',
            cdr3nt_a_col = 'cdr3a_nucseq',
            vgene_a_col = 'va',
            jgene_a_col = 'ja',
            cdr3aa_b_col = 'cdr3b',
            cdr3nt_b_col = 'cdr3b_nucseq',
            vgene_b_col = 'vb',
            jgene_b_col = 'jb',
            remove_nonfunctional=True
            ):
        
        col_remap = {
            cdr3nt_a_col:"cdr3a",
            cdr3aa_a_col:"cdr3a_nucseq",
            vgene_a_col:"va",
            jgene_a_col:"ja",
            cdr3nt_b_col:"cdr3b",
            cdr3aa_b_col:"cdr3b_nucseq",
            vgene_b_col:"vb",
            jgene_b_col:"jb",
            }

        # Remove ambiguous CDR3 amino acid sequences
        print("Remove ambiguous CDR3 amino acid sequences")
        self.data = self.data[self.data.cdr3a.apply(lambda cdr3: _is_cdr3(cdr3)) &
                              self.data.cdr3b.apply(lambda cdr3: _is_cdr3(cdr3))]

        # Remove ORF and non-functional genes
        print("Parsing V/J genes")
        if remove_nonfunctional:
            functional = IMGT[IMGT['fct'].isin(['F','(F)','[F]'])]
            self.data = self.data[self.data.va.isin(functional.imgt_allele_name) & 
                                  self.data.vb.isin(functional.imgt_allele_name)]
        else:
            # make sure all V genes have IMGT annotation
            self.data = self.data[self.data.va.isin(IMGT.imgt_allele_name) & 
                                  self.data.vb.isin(IMGT.imgt_allele_name)]
        
        tcra = self.data[['cdr3a','cdr3a_nucseq','va','ja']]
        tcrb = self.data[['cdr3b','cdr3b_nucseq','vb','jb']]
        tcra.columns = ['junction_aa','junction','v_call','j_call']
        tcrb.columns = ['junction_aa','junction','v_call','j_call']
        tcra = format_junction(tcra)
        tcrb = format_junction(tcrb)
        self.data['cdr3a_nucseq'] = tcra['junction']
        self.data['cdr3b_nucseq'] = tcrb['junction']
        self.data = self.data.dropna(subset=["cdr3a","cdr3b","cdr3a_nucseq","cdr3b_nucseq","va","ja","vb","jb"])

        # # Apply format_junction directly to the relevant columns
        # self.data[['cdr3a_nucseq', 'cdr3b_nucseq']] = self.data[['cdr3a', 'cdr3a_nucseq', 'va', 'ja', 'cdr3b', 'cdr3b_nucseq', 'vb', 'jb']].apply(
        #     lambda x: format_junction(x[['cdr3a', 'cdr3a_nucseq', 'va', 'ja']],'cdr3a','cdr3a_nucseq','va','ja')['junction']
        #     if x.name == 'cdr3a_nucseq' else format_junction(x[['cdr3b', 'cdr3b_nucseq', 'vb', 'jb']],'cdr3b', 'cdr3b_nucseq', 'vb', 'jb')['junction'], axis=1
        # )

        return self.data


    def airr_to_tcrdist_paired(self):
        '''
        Convert AIRR data to paired tcrdist format.
        '''
        # Splitting the dataframe into TRA and TRB
        alpha_data = self.data[self.data['locus'] == 'TRA'].copy()
        beta_data = self.data[self.data['locus'] == 'TRB'].copy()
        # Renaming columns for merging
        alpha_data = alpha_data.rename(columns={'v_call': 'va', 'j_call': 'ja', 'junction_aa': 'cdr3a', 'junction': 'cdr3a_nucseq'})
        beta_data = beta_data.rename(columns={'v_call': 'vb', 'j_call': 'jb', 'junction_aa': 'cdr3b', 'junction': 'cdr3b_nucseq'})
        # alpha_data = alpha_data.drop(columns=['locus'])
        # beta_data = beta_data.drop(columns=['locus'])
        
        possible_id_cols = [['cell_id', 'clone_id'], ['cell_id'], ['clone_id']]
        id_cols = next((cols for cols in possible_id_cols if all(col in self.data.columns for col in cols)), None)
        if id_cols is None:
            raise ValueError("No valid id columns found in the dataframe. Paired data must include clone_id and/or cell_id.")
        
        alpha_cols = ['va', 'ja', 'cdr3a', 'cdr3a_nucseq'] + id_cols
        beta_cols = ['vb', 'jb', 'cdr3b', 'cdr3b_nucseq'] + id_cols
        alpha_data = alpha_data[alpha_cols]
        beta_data = beta_data[beta_cols]
        
        merged = pd.merge(alpha_data, beta_data, on=id_cols, how='inner')
        merged = merged.dropna()
        return merged


import numpy as np
import pandas as pd
import parmap
import faiss
import time
import json
import os

from scipy.stats import hypergeom
from multiprocessing import Pool, cpu_count
from typing import Union
from functools import reduce

from .encoding import TCRDistEncoder
from .indexing import IvfIndex, FlatIndex
from .background import BackgroundModel
from .constants.parsing import check_formatting

import faiss
import pandas as pd
import numpy as np
import networkx as nx
import leidenalg as la
import igraph as ig

from scipy.spatial.distance import pdist, squareform
from scipy.sparse import coo_matrix, lil_matrix, csr_matrix
from scipy.stats import hypergeom
from timeit import default_timer as timer
from snetcr.encoding import TCRDistEncoder
from snetcr.constants.preprocessing import format_chain
from snetcr.indexing import FlatIndex, IvfIndex
from snetcr.background import BackgroundModel
from snetcr.repertoire import Repertoire

class SneTcrResult:
    '''
    Class for storing the results of a neighborhood analysis.

    Attributes:
        data (pd.DataFrame): The data containing the results of the neighborhood analysis.
        vecs (np.ndarray, optional): The vector encodings of the data. Defaults to None.

    Methods:
        __init__(self, sne, vecs=None): Initializes a SneTcrResult object.
        __repr__(self): Returns a string representation of the SneTcrResult object.
        __str__(self): Returns a string representation of the SneTcrResult object.
        __len__(self): Returns the length of the SneTcrResult object.
        __getitem__(self, n): Returns the n-th item of the SneTcrResult object.
        to_df(self): Converts the SneTcrResult object to a pandas DataFrame.
        get_vecs(self): Computes the vector encodings of the data.
        get_clusters(self, r=96, chain='AB'): Gets the different SNE clusters.

    '''

    def __init__(
            self, 
            sne,
            chain='AB',
            vecs=None
            ) -> None:
        '''
        Initializes a SneTcrResult object.

        Parameters:
            sne (pd.DataFrame): The data containing the results of the neighborhood analysis.
            vecs (np.ndarray, optional): The vector encodings of the data. Defaults to None.
        '''
        
        self.data = sne.reset_index(drop=True)
        self.chain = chain
        self.vecs = vecs

    def __repr__(self) -> str:
        '''
        Returns a string representation of the SneTcrResult object.
        '''
        return f"SneTcrResult"
    
    def __str__(self) -> str:
        '''
        Returns a string representation of the SneTcrResult object.
        '''
        return f"SneTcrResult"

    def __len__(self) -> int:
        '''
        Returns the length of the SneTcrResult object.
        '''
        return len(self.data)
    
    def __getitem__(self, n):
        '''
        Returns the n-th item of the SneTcrResult object.
        '''
        return self.data[n]
    
    def to_df(self) -> pd.DataFrame:
        '''
        Converts the SneTcrResult object to a pandas DataFrame.

        Returns:
            pd.DataFrame: The SneTcrResult object as a pandas DataFrame.
        '''
        return self.data
    
    def get_vecs(self):
        '''
        Computes the vector encodings of the data.
        '''
        encoder = TCRDistEncoder(aa_dim=16, chain=self.chain).fit()
        self.vecs = encoder.transform(self.data)

    def _index_based_graph(self, r, significant=True):

        if significant:
            sign = self.data[self.data['evalue'] < .05]

        sign_vecs = self.vecs[sign.index]

        index = faiss.IndexFlatL2(self.vecs.shape[1])
        index.add(self.vecs)

        lims, D, I = index.range_search(sign_vecs, thresh=r)
        edges = convert_range_search_output(lims, D, I)
        edges = [i[0] for i in edges if len(set(i[0])) > 1]
        nodes = set(list(sign.index) + list(I))

        return edges, nodes
    
    def _matrix_based_graph(self, r, significant=True):

        if significant:
            sign = self.data[self.data['evalue'] < .05]

        dm = compute_sparse_distance_matrix(
            tcrs=sign, 
            chain=self.chain, 
            organism='human', 
            d=r,
            m=16,
            vecs=self.vecs[sign.index]
            )
        
        non_zero_values = dm.data
        row_indices, col_indices = dm.nonzero()

        mask = (non_zero_values >= 0)
        filtered_row_indices = row_indices[mask]
        filtered_col_indices = col_indices[mask]
        ids = np.column_stack((filtered_row_indices, filtered_col_indices))

        nodes = sign.reset_index(drop=True).index.values
        edges = [(nodes[i[0]], nodes[i[1]]) for i in ids]   
        
        return edges, nodes

    def get_clusters(self, r=96, significant=True, periphery=False):
        '''
        Gets the different SNE clusters.
        Builds a network from sequence neighbors (< r) and partitions
        the graph using Leiden clustering.

        Suggested thresholds for r:
            - paired chain ('AB'): 96
            - single chain ('A' or 'B'): 24

        Parameters:
            r (int): 96
                Threshold for defining a neighbor.
            chain (str): 'AB'
                TCR chain.

        Returns:
            clusters (dict): Dictionary with cluster assignments per TCR. The keys correspond to the TCR indices in the neighbor analysis dataframe.

        Also adds a cluster column to the dataframe.
        '''
        
        if self.vecs is None:
            self.get_vecs()

        # if significant:
        #     sign = self.data[self.data['evalue'] < .05]

        # dm = compute_sparse_distance_matrix(
        #     tcrs=sign, 
        #     chain=self.chain, 
        #     organism='human', 
        #     d=r,
        #     m=16,
        #     vecs=self.vecs[sign.index]
        #     )
        
        # non_zero_values = dm.data
        # row_indices, col_indices = dm.nonzero()

        # mask = (non_zero_values >= 0)
        # filtered_row_indices = row_indices[mask]
        # filtered_col_indices = col_indices[mask]
        # ids = np.column_stack((filtered_row_indices, filtered_col_indices))

        # dm = squareform(pdist(self.vecs, metric='euclidean')**2)
        # np.fill_diagonal(dm, -1)
        # ids = np.argwhere((dm <= r) & (dm >= 0))

        if periphery:
            edges, nodes = self._index_based_graph(r)
        else:
            edges, nodes = self._matrix_based_graph(r)
        
        # nodes = sign.reset_index(drop=True).index.values
        # edges = [(nodes[i[0]], nodes[i[1]]) for i in ids]
        G = nx.Graph()
        G.add_nodes_from(list(nodes))
        G.add_edges_from(edges)

        if isinstance(G, (nx.Graph, nx.DiGraph)):
            G = ig.Graph.from_networkx(G)
        partition = la.find_partition(G, la.ModularityVertexPartition)
        cluster_lists = [[list(nodes)[node] for node in c] for c in list(partition)]
        clusters = {int(j): n for n, i in enumerate(cluster_lists) for j in i}
        self.data['cluster'] = self.data.index.map(clusters)
        self.data['cluster'] = self.data['cluster'].fillna(-1).astype(int)

        return clusters

def neighbor_analysis(tcrs, chain: str, organism: str, radius: Union[float,int], vecs=None, background=None, encoder=None, fgindex=None, bgindex=None, depth: int=10):
    '''
    Perform a neighborhood analysis on a set of TCRs. TCR neighborhood analysis compares the
    distribution of sequence neighbors within a fixed distance radius against a background dataset.
    Neighborhood enrichment is determined by comparing the neighbor count of a TCR in the sample
    compared to the neighbor count of the same TCR in the background dataset.

    Suggestions for the choice of 'radius':
        - single chain: 12.5 (strict), 24.5 (flexible)
        - paired chain: 96.5

    Parameters:
        tcrs
            The input data provided as a pandas.DataFrame.
        chain (str): 'AB'
            TCR chain.
        organism (str): human
            Species from which the data originates.
        radius (float or int)
             Threshold for defining a neighbor.
        vecs: None
            Pre-computed vecTCRdist encodings.
        background: None
            Custom background.
        encoder : None
            Custom encoder. Accepts TCRDistEncoder only.
        fgindex: None
            Custom index for foreground TCRs.
        bgindex: None
            Custom index for background TCRs.
        depth (int): 10
            Depth of the background data set. This is a factor of the size of the input data.

    Returns:
        SneTcrResult
    '''

    chain = format_chain(chain)

    if len(chain) == 1:

        if vecs is None:
            if encoder is None:
                encoder = TCRDistEncoder(aa_dim=8,organism=organism,chain=chain).fit()
            print('Encoding TCRs.')
            vecs = encoder.transform(tcrs)

        print('Computing neighbor distribution.')
        if fgindex is None:
            fgindex = faiss.IndexFlatL2(vecs.shape[1])
            fgindex.add(vecs)
        
        lims, D, I = fgindex.range_search(vecs, radius)
        nbrcounts = neighbor_distr_from_lims(lims)
        nonzero = np.where(nbrcounts > 0)[0]
        del fgindex

        if background is not None:
            print('Background data provided, skipping background generation.')
        else:
            background = BackgroundModel(tcrs, 10)
            bgdata = background.shuffle(chain=chain)
        
        bgvecs = encoder.transform(bgdata)

        if bgindex is None:
            bgindex = faiss.IndexFlatL2(vecs.shape[1])
            bgindex.add(bgvecs)

        lims, D, I = bgindex.range_search(vecs[nonzero], 12.5)
        del bgindex

        estimated = neighbor_distr_from_lims(lims)

        presne = tcrs.loc[nonzero]
        presne['foreground_neighbors'] = nbrcounts[nonzero]
        presne['background_neighbors'] = estimated

        sne = _hypergeometric(presne, tcrs.shape[0], bgdata.shape[0])

    elif len(chain) > 1:

        # Make sure the data is in the paired format
        airr_cols = ['v_call','j_call','junction_aa','junction']
        if all([i in tcrs.columns for i in airr_cols]):
            rep = Repertoire(tcrs)
            tcrs = rep.airr_to_tcrdist_paired()
        else:
            pass
        
        print('Encoding TCRs')
        if vecs is None:
            if encoder is None:
                encoder = TCRDistEncoder(aa_dim=8,organism=organism,chain=chain).fit()
            vecs = encoder.transform(tcrs)
        
        # Get the neighbor distribution in the sample
        print(f'Computing neighbor distribution for {vecs.shape[0]} TCRs in sample.')
        fg_results = get_foreground_nbr_counts(vecs, radius)

        if background is None:
            # Create a background data set
            print(f'Creating background data set with {tcrs.shape[0]*depth} TCRs')
            bgmodel = BackgroundModel(repertoire=tcrs, factor=depth)
            background = bgmodel.shuffle(chain=chain)

        # Get the encodings for both chains separately in foreground and background
        print('Estimating expected neighbor distribution.')
        avecs, bvecs = encoder.transform(tcrs, split_ab=True)
        avecsbg, bvecsbg = encoder.transform(background, split_ab=True)
        bg_ab_counts = get_background_nbr_counts(avecs,bvecs,avecsbg,bvecsbg,radius)

        # Combine all results and compute the neighborhood p-values
        sne = compute_neighborhood_pvalues(tcrs, fg_results, bg_ab_counts, background.shape[0])

    return SneTcrResult(sne=sne, chain=chain)

def neighbor_distr_from_lims(lims):
    '''
    Extract the number of neighbors for each query from lims.
    '''
    num_queries = len(lims) - 1
    result = np.zeros(num_queries, dtype=int)
    for i in range(num_queries):
        start = lims[i]
        end = lims[i + 1]
        result[i] = end - start
    return result

def convert_range_search_output(lims, D, I, offset=0):
    '''
    Convert the output of a range search to a list of tuples.
    '''
    # Calculate the number of queries
    num_queries = len(lims) - 1
    # Vectorize the construction of result pairs
    result_indices = []
    result_distances = []
    for i in range(num_queries):
        start = lims[i]
        end = lims[i + 1]
        query_indices = np.full(end - start, i)
        neighbor_indices = I[start:end] + offset
        distances = D[start:end]
        result_indices.extend(zip(query_indices, neighbor_indices))
        result_distances.extend(distances)
    result = list(zip(result_indices, result_distances))
    return result
    
def compute_sparse_distance_matrix(tcrs, chain, organism, exact=True, d=96.5, m=8, encoder=None, vecs=None):
    '''
    Compute the sparse distance matrix for a set of TCRs.
    '''
    
    t0 = timer()

    if encoder == None:
        encoder = TCRDistEncoder(aa_dim=m,organism=organism,chain=chain).fit()

    if vecs is None:
        # Encode the TCRs
        start = timer()
        vecs = encoder.transform(tcrs).astype(np.float32)
        print(f'Encoding TCRs took {timer()-start:.2f}s')

    if exact:
        # Flat index will ensure 'exact' search
        # However, this is still an approximation of the true TCRdist
        start = timer()
        index = FlatIndex(encoder=encoder)
        index.add(tcrs, vecs)
        print(f'Building the index took {timer()-start:.2f}s')
    else:
        # IVF index
        start = timer()
        k = round(len(vecs)/1000)
        n = round(k/20)+2
        if n > k:
            n = k
        index = IvfIndex(encoder=encoder, n_centroids=k, n_probe=n)
        index.add(tcrs, vecs)
        print(f'Building the index took {timer()-start:.2f}s')

    # Search index
    start = timer()
    lims, D, I = index.idx.range_search(vecs, thresh=d)
    print(f'Searching index within radius {d} took {timer()-start:.2f}s')

    # Format the results.
    start = timer() 
    res = convert_range_search_output(lims, D, I)

    # Determine the size of the matrix
    max_index = len(tcrs)
    base = [([i,i],0) for i in range(max_index)]
    res += base
    res = list(set((tuple(indices), distance-1) for indices, distance in res))

    # Initialize lists to hold row indices, column indices, and data values
    rows, cols, values = [], [], []

    # Populate the lists with the given distances
    for indices, distance in res:
        i, j = indices
        rows.append(i)
        cols.append(j)
        values.append(distance)
        if i != j:  # Ensure the matrix is symmetric
            rows.append(j)
            cols.append(i)
            values.append(distance)

    # Convert the lists to a sparse matrix using coo_matrix
    dm = coo_matrix((values, (rows, cols)), shape=(max_index, max_index), dtype=np.float64)
    print(f'Building the sparse matrix took {timer()-start:.2f}s')
    print(f'Total time to compute distance matrix: {timer()-t0:.2f}s')

    return dm

def find_neighbors(tcrs, chain, radius, organism='human', vecs=None, encoder:TCRDistEncoder=None, index=None):
    '''
    Suggested radii:
        - single chain: 24.5
        - paired chain: 96.5
    '''

    if vecs is None:
        if encoder is None:
            encoder = TCRDistEncoder(aa_dim=8,organism=organism,chain=chain).fit()
        vecs = encoder.transform(tcrs)

    if index is None:
        index = faiss.IndexFlatL2(vecs.shape[1])
        index.add(vecs)
    
    lims, D, I = index.range_search(vecs,radius)

    return neighbor_distr_from_lims(lims)

def _hypergeometric(
        data,
        fg_size,
        bg_size,
        pseudocount=0
        ):
    '''
    Compute p-values for based on foreground and background neighbor counts
    using the hypergeometric distribution.

    NOTE:
    SciPy's hypergeometric sf function uses the boost distribution implementation.
    Integer values in the boost distribution use the C unsigned integer type, which for most
    compilers is 32 bits. Any values exceedingn 32 bits will prompt either nan or strange
    p-values.
    '''

    # Hypergeometric testing
    N = fg_size - 1
    M = bg_size + N
    if M >= 2**32:
        import warnings
        warnings.warn(f"Population size exceeds {2**32-1} (32 bits).\n p-values may be miscalculated.")
    data['pvalue'] = data.apply(
        lambda x: hypergeom.sf(
            x['foreground_neighbors']-1, 
            M, 
            x['foreground_neighbors'] + x['background_neighbors'] + pseudocount, 
            N
            ),
        axis=1
        )
    data['evalue'] = data['pvalue'] * fg_size
    
    return data.sort_values(by='evalue')

def get_foreground_nbr_counts(
        fg_vecs,
        radius=96.5,
):
    '''
    Get the faiss data (lims,D,I) for range_search of tcrs against self

    so it will include self-distances

    tcrs and bg_tcrs should already have been filtered
    '''
    qvecs = fg_vecs # could optionally split into batches

    print('start IndexFlatL2 range search', qvecs.shape, fg_vecs.shape)
    start = timer()
    idx = faiss.IndexFlatL2(fg_vecs.shape[1])
    idx.add(fg_vecs)
    print(type(qvecs), type(radius))
    lims, D, I = idx.range_search(qvecs, radius)
    print(f'IndexFlatL2 range search took {timer()-start:.2f}')
    return {'lims':lims, 'D':D, 'I':I}

def get_background_nbr_counts(
        fg_avecs,
        fg_bvecs,
        bg_avecs,
        bg_bvecs,
        maxdist=96,
):
    ''' Compute the background paired tcrdist distribution by taking the
    convolution of the alpha and beta single-chain tcrdist distributions.
    The effective number of paired background comparisons is len(bg_tcrs)**2

    returns an integer-valued numpy array of shape (num_fg_tcrs, maxdist+1)

    histogram bin-size is 1.0

    first bin is (-0.5, 0.5), last bin is (maxdist-0.5, maxdist+0.5)

    tcrs and bg_tcrs should already have been filtered

    (see phil_functions.compute_background_paired_tcrdist_distributions)
    '''

    start = timer()
    ab_counts = compute_background_paired_tcrdist_distributions(
        fg_avecs, fg_bvecs, bg_avecs, bg_bvecs, maxdist)
    print(f'paired distribution calc took {timer()-start:.2f}')

    return ab_counts

def compute_background_single_tcrdist_distributions(
        fg_vecs,
        bg_vecs,
        maxdist,
        rowstep = 5000,
        colstep = 50000,
):
    ''' Compute the histogram of distances to bg_vecs for each vector in fg_vecs

    returns an integer-valued numpy array of shape (num_fg_tcrs, maxdist+1)

    histogram bin-size is 1.0

    first bin is (-0.5, 0.5), last bin is (maxdist-0.5, maxdist+0.5)

    for alphabeta tcr clumping, I use maxdist=96
    '''
    dim = fg_vecs.shape[1]
    assert dim == bg_vecs.shape[1]
    num_fg = fg_vecs.shape[0]
    num_bg = bg_vecs.shape[0]
    maxdist = int(maxdist+0.1) # confirm int
    print('compute_background_single_tcrdist_distributions: '
          f'dim= {dim} maxdist= {maxdist}')
    #maxdist_float = maxdist + 0.5
    dist_counts = np.zeros((num_fg, maxdist+1), dtype=int)

    nrows = (num_fg-1)//rowstep + 1
    ncols = (num_bg-1)//colstep + 1

    # initialize distance storage
    dists = np.zeros((rowstep*colstep,), dtype=np.float32) # 1D array

    start0 = timer()
    for ii in range(nrows):
        ii_start = ii*rowstep
        for jj in range(ncols):
            jj_start = jj*colstep
            xq = fg_vecs[ii_start:ii_start+rowstep] # faiss terminology here
            xb = bg_vecs[jj_start:jj_start+colstep]
            nq = xq.shape[0] # Num Query
            nb = xb.shape[0] # Num dataBase (?)

            start = timer()
            faiss.pairwise_L2sqr(dim, nq, faiss.swig_ptr(xq), nb, faiss.swig_ptr(xb),
                                 faiss.swig_ptr(dists))
            disttime = timer()-start
            #print(disttime)
            start += disttime

            # now fill counts
            for iq in range(nq):
                dist_counts[ii_start+iq,:] += np.histogram(
                    dists[iq*nb:(iq+1)*nb], bins = maxdist+1,
                    range = (-0.5, maxdist+0.5))[0]

            counttime = timer()-start

            #print(f'ij {ii} {jj} {disttime:.2f} {counttime:.2f} {timer()-start0:.2f}')

    return dist_counts



def compute_background_paired_tcrdist_distributions(
        fg_avecs,
        fg_bvecs,
        bg_avecs,
        bg_bvecs,
        maxdist,
        rowstep = 5000,
        colstep = 50000,
):
    ''' Compute the background paired tcrdist distribution by taking the
    convolution of the alpha and beta single-chain tcrdist distributions

    returns an integer-valued numpy array of shape (num_fg_tcrs, maxdist+1)

    histogram bin-size is 1.0

    first bin is (-0.5, 0.5), last bin is (maxdist-0.5, maxdist+0.5)
    '''

    num_fg_tcrs = fg_avecs.shape[0]
    num_bg_tcrs = bg_avecs.shape[0]
    assert num_fg_tcrs == fg_bvecs.shape[0]
    assert num_bg_tcrs == bg_bvecs.shape[0]

    acounts = compute_background_single_tcrdist_distributions(
        fg_avecs, bg_avecs, maxdist, rowstep=rowstep, colstep=colstep)

    bcounts = compute_background_single_tcrdist_distributions(
        fg_bvecs, bg_bvecs, maxdist, rowstep=rowstep, colstep=colstep)

    abcounts = np.zeros((num_fg_tcrs, int(maxdist+1)), dtype=int)

    assert acounts.shape == bcounts.shape == abcounts.shape

    for d in range(int(maxdist+1)):
        for adist in range(d+1):
            abcounts[:,d] += acounts[:,adist] * bcounts[:,d-adist]

    return abcounts

def compute_neighborhood_pvalues(
        tcrs,
        fg_results,
        bg_ab_counts,
        num_bg_tcrs,
        radii = [24, 48, 72, 96],
        min_nbrs = 2, # tcr_clumping uses 1?
        pseudocount = 0.25,
        evalue_threshold = 1,
):
    ''' Compute pvalues and simple-bonferroni corrected "evalues"
    for observed foreground neighbor numbers

    Right now this is using poisson but we could shift to hypergeometric
    I think for the paired setting where the effective number of background comparisons
    is very large the two should give pretty similar results
    '''
    from scipy.stats import poisson

    # get background counts at radii
    maxdist = max(radii)
    assert bg_ab_counts.shape[1] == maxdist+1
    bg_counts = np.cumsum(bg_ab_counts, axis=1)[:, radii]

    # get foreground counts at radii
    num_fg_tcrs = tcrs.shape[0]
    lims = fg_results['lims']
    D = fg_results['D']
    assert num_fg_tcrs == lims.shape[0]-1
    fg_counts = np.zeros((num_fg_tcrs, len(radii)), dtype=int)
    for ii, r in enumerate(radii):
        fg_counts[:,ii] = np.add.reduceat((D<=r+.5), lims[:-1].astype(int))

    fg_counts -= 1 # exclude self nbrs

    assert fg_counts.shape == bg_counts.shape == (num_fg_tcrs, len(radii))

    # "rates" are the expected number of neighbors based on the background counts
    # we divide background counts by num_bg_tcrs**2 (since that's the effective number
    # of background paired comparisons) to get the probability of seeing a neighbor
    # at a given radius, then we multiply by num_fg_tcrs to get the expected number
    # of neighbors.
    # rates.shape: (num_fg_tcrs, len(radii))
    #
    rates = (np.maximum(bg_counts, pseudocount) *
             (num_fg_tcrs/(num_bg_tcrs*num_bg_tcrs)))

    dfl = []
    for ii, (counts,rates) in enumerate(zip(fg_counts, rates)):
        for jj, (count,rate) in enumerate(zip(counts,rates)):
            if count >= min_nbrs:
                pvalue = poisson.sf(count-1, rate)
                evalue = pvalue*(len(radii)*num_fg_tcrs)
                if evalue <= evalue_threshold:
                    dfl.append(dict(
                        pvalue=pvalue,
                        evalue=evalue,
                        tcr_index=ii,
                        radius=radii[jj],
                        num_nbrs=count,
                        expected_num_nbrs=rate,
                        bg_nbrs=rate*num_bg_tcrs*num_bg_tcrs/num_fg_tcrs,
                    ))

    results = pd.DataFrame(dfl)
    if results.shape[0]:
        results = results.join(tcrs, on='tcr_index').sort_values('evalue')

    return results

def above_threshold(df, row, t):
    for column in df.columns:
        if row[column] > t:
            return column
    return None

def tcr_dict_to_df(neighbor_counts, cutoff:int=0, add_counts:bool=True):
    '''
    Helper function to convert dictionary of joined V_CDR3
    representations into a DataFrame format.

    Parameters
    ----------
    neighbor_counts
        Dictionary of TCRs (format: {V_CDR3 : n}).
        This corresponds to the output of the radius search function in NeighborEnrichment
    cutoff, int
        Filters out TCRs with <= cutoff neighbors. Default is 0
    add_counts, bool
        Add neighbor counts. Default is True.
    '''
    neighbor_counts = {i:neighbor_counts[i] for i in neighbor_counts if neighbor_counts[i] >= cutoff}
    vgenes = [i.split('_')[0] for i in list(neighbor_counts.keys())]
    cdr3aa = [i.split('_')[1] for i in list(neighbor_counts.keys())]
    if not add_counts:
        return pd.DataFrame({'v_call':vgenes, 'junction_aa':cdr3aa})
    else:
        counts = list(neighbor_counts.values())
        return pd.DataFrame({'v_call':vgenes, 'junction_aa':cdr3aa, 'neighbors':counts})

def assign_ids(query, lim):
    '''
    Match neighbor counts and indices.

    Parameters
    ----------
    query
        TCRs or CDR3 sequences for which the number of neighbors on the index
        was calculated.
    lim
        Faiss range search outcome that defines the number of retrieved neighbors
        for every queried vector.
    '''
    if isinstance(query, list):
        return dict([(j, int(lim[i+1]-lim[i])-1) for i,j in enumerate(query)])
    elif isinstance(query, pd.DataFrame):
        indices = (query['v_call'] + "_" + query['junction_aa']).to_list()
        return dict([(j, int(lim[i+1]-lim[i])-1) for i,j in enumerate(indices)])

def index_neighbors(query, r, index):
    '''
    Compute the number of neighbors on the index that are found within 
    a radius r of each query TCR. Returns a dict of query sequences
    and their respective neighbor count.

    Parameters
    ----------
    query
        TCRs or CDR3 sequences for which the number of neighbors on the index
        should be calculated.
    r
        Radius parameter that defines the edge of the TCR neighborhood.
    index
        Index on which the search should be performed. The index needs to be
        trained and should contain vectors to query against.
    '''
    X = index.encoder.fit_transform(query)
    X = X.astype(np.float32)
    lim, D, I = index.idx.range_search(X, thresh=r)
    return assign_ids(query=index.encoder.tcrs, lim=lim)

def neighbor_retrieval(query, encoder=None, background=None, index_type="exact", d=24.5, drop_zero_neighbor=False):
    """
    Retrieve neighboring T-cell receptor (TCR) sequences from a background set based on a similarity threshold.

    This function takes a query set of TCR sequences and a background set (or uses the query set itself if none provided),
    and retrieves neighboring TCR sequences from the background set based on their similarity to the query set,
    using a specified encoding method.

    Parameters:
        query (pandas.DataFrame): A DataFrame containing TCR sequences to be used as the query set.
        encoder (Optional[TCRDistEncoder]): An instance of TCRDistEncoder for encoding TCR sequences.
            If not provided, a default TCRDistEncoder will be used.
        background (Optional[pandas.DataFrame]): A DataFrame containing TCR sequences to be used as the background set.
            If not provided, the query set will be used as the background set.
        d (float): The similarity threshold. TCR sequences with a similarity score below this threshold will be excluded.
        drop_zero_neighbor (bool): If True, TCRs with no neighbors within the specified threshold will be excluded from the result.

    Returns:
        dict: A dictionary containing information about the neighboring TCR sequences for each query TCR.
            The keys of the dictionary are tuples representing the (v_call, junction_aa) of the query TCRs.
            The values are tuples containing:
                - The number of neighbors found within the threshold (nneigh).
                - List of indices of neighboring TCRs in the background set (idx_ids).
                - List of similarity distances between query TCRs and their neighbors (idx_dist).
                - List of neighboring TCR sequences (idx_tcr).

    Note:
        - TCRDistEncoder is used to encode TCR sequences into numerical vectors for similarity comparison.
        - The function performs range search on the encoded TCR sequences to find neighbors.
        - If drop_zero_neighbor is True, TCRs with no neighbors within the threshold are excluded from the result.

    Example:
        query_df = pd.DataFrame({'v_call': ['TRBV6-1*01', 'TRBV7-2*01'], 'junction_aa': ['CASSSPSGAGALHF', 'CASSSRDLPTEAFF']})
        background_df = pd.DataFrame({'v_call': ['TRBV3-1*01', 'TRBV3-1*01'], 'junction_aa': ['CASSQPEPTSFERANTGELFF', 'CASSQPRISTSGGNSTDTQYF']})
        neighbor_dict = neighbor_retrieval(query_df, background_df, d=24.5)
    """
    # If none provided, use default TCRDistEncoder
    if encoder == None:
        encoder = TCRDistEncoder(full_tcr=True, aa_dim=16).fit()
    # Encode query TCRs and create index
    print("encoding query sequences")
    query_vecs = encoder.transform(query)
    # If the user provides an index, we can skip the index building step
    if isinstance(background, (FlatIndex, IvfIndex)):
        assert background.n > 0, "Empty index provided. Please provide a fitted index or a data frame with TCRs."
        index = background
    else:
        # If none provided, search query TCRs against itself
        if background is None:
            background = query
        if index_type == "exact":
            print("building flat index")
            index = FlatIndex(encoder=encoder)
        else:
            print("building IVF index")
            n_centr = int(background.shape[0] / 1000)
            n_probe = int(n_centr / 50)
            index = IvfIndex(encoder=encoder, n_centroids=n_centr, n_probe=n_probe)
        index.add(background)
    # Perform search on the index
    print("searching for neighbors")
    lims, D, I = index.idx.range_search(query_vecs.astype(np.float32), thresh=d)
    # Build the neighbor dictionary
    print("Combining results")
    neighbor_dict = {}
    for n, i in enumerate(query.iterrows()):
        nneigh = lims[n+1]-lims[n]
        idx_ids = list(I[lims[n]:lims[n+1]])
        idx_dist = list(D[lims[n]:lims[n+1]])
        idx_tcr = [index.ids[j] for j in idx_ids]
        neighbor_dict[(i[1].v_call, i[1].junction_aa)] = (nneigh, idx_ids, idx_dist, idx_tcr)
    # Drop out TCRs with no neighbors <= d if True
    if drop_zero_neighbor:
        neighbor_dict = {i:neighbor_dict[i] for i in neighbor_dict if neighbor_dict[i][0] > 0}

    return neighbor_dict

def neighbor_dict_to_df(neighbor_dict):
    """
    Convert a neighbor dictionary to a pandas DataFrame representing an adjacency list.

    This function takes a neighbor dictionary, which contains information about neighboring TCR sequences for each query TCR,
    and converts it into a pandas DataFrame that represents an adjacency list suitable for network analysis.

    Parameters:
        neighbor_dict (dict): A dictionary containing information about neighboring TCR sequences.
            The keys are tuples representing the (v_call, junction_aa) of query TCRs.
            The values are tuples containing:
                - The number of neighbors found (nneigh).
                - List of indices of neighboring TCRs (idx_ids).
                - List of similarity distances between query TCRs and their neighbors (idx_dist).
                - List of neighboring TCR sequences (idx_tcr).

    Returns:
        pandas.DataFrame: A DataFrame representing an adjacency list with columns:
            - 'idx': Index of the neighboring TCR.
            - 'tcrdist': Similarity distance between the query TCR and the neighbor.
            - 'target': TCR sequence of the neighbor.
            - 'source': TCR sequence of the query TCR (formatted as 'v_call_junction_aa').
            - 'target_v_call': V gene of the neighbor.
            - 'target_junction_aa': CDR3 of the neighbor.
            - 'source_v_call': V gene of the query TCR.
            - 'source_junction_aa': CDR3 of the query TCR.

    Example:
        neighbor_dict = {('V1', 'ACDE'): (2, [0, 1], [0.1, 0.2], ['V2_ACDF', 'V3_EFGH'])}
        adjacency_df = neighbor_dict_to_df(neighbor_dict)
    """
    adj_list = []
    # For every TCR in the neighbor dictionary, create an adjacency matrix
    for i in neighbor_dict:
        n = neighbor_dict[i][0] # number of neighbors
        df = pd.DataFrame([tuple(j[x] for j in neighbor_dict[i][1:]) for x in range(n)])
        df.columns = ["idx", "tcrdist", "background"]
        df["query"] = "_".join(i) # source TCR
        adj_list.append(df)
    # Concatenate all adjacency lists into one
    adj_list = pd.concat(adj_list)
    # Annotate V gene and CDR3
    adj_list[["background_v_call", "background_junction_aa"]] = adj_list["background"].str.split("_", expand=True)
    adj_list[["query_v_call", "query_junction_aa"]] = adj_list["query"].str.split("_", expand=True)
    
    return adj_list




# LEGACY
class NeighborEnrichment():
    def __init__(
        self,
        repertoire:Union[pd.DataFrame, list],
        encoder:TCRDistEncoder=None,
        organism='human',
        background=None,
        exact=True,
        k=None,
        n=None,
        num_workers=1,
        add_pseudocount=False
        # custom_background=None, -> add this functionality back in later
        ):
        '''
        Class for calculating and statistically evaluating the neighbor count of 
        TCRs or CDR3 sequences.

        Parameters
        ----------
        repertoire: Union[pd.DataFrame, list]
            Repertoire of interest. This can be a list of CDR3 sequences
            as well as a pd.DataFrame containing at least V and CDR3
            information. Column names should satisfy AIRR data conventions.
        encoder: Union[TCRDistEncoder]
            Transformer function used to embed sequences into vector space.
        
        '''
        self.repertoire = repertoire

        # self.background = custom_background -> add this functionality back in later
        if encoder == None:
            self.encoder = TCRDistEncoder(aa_dim=8, full_tcr=True).fit()
        else:
            self.encoder = encoder

        # before proceeding, check whether all data fields are formatted correctly
        if encoder.organism == "human":
            check_formatting(self.repertoire)
        else:
            pass
        
        self.organism = organism
        self.rsize = len(repertoire)
        # self.r = radius
        self.background = background
        self.nbr_counts = None
        self.bg_index = None
        self.num_workers = num_workers

        self.encoder.fit()

        if exact:
            self.fg_index = FlatIndex(encoder=self.encoder)
        else:
            if k is None:
                k = np.round(self.rsize/1000,0)
            if n is None:
                n = np.round(k/30,0)
            self.fg_index = IvfIndex(encoder=self.encoder, n_centroids=int(k), n_probe=int(n))

        if add_pseudocount:
            self.pseudocount = 1
        else:
            self.pseudocount = 0


    def fixed_radius_neighbors(self, radius:Union[int,float]=12.5):
        '''
        Find the number of neighbors within a repertoire of interest.

        Parameters
        ----------
        radius: [int, float]
            Radius (in TCRdist units) around which to search for neighbors.
        k
            Number of centroids for fitting faiss.IndexIVFFlat.
        n
            Number of cells to probe when using the faiss.IndexIVFFlat.
        '''
        self.fg_index.add(self.repertoire)
        self.r = radius
        self.nbr_counts = tcr_dict_to_df(
            index_neighbors(query=self.repertoire, index=self.fg_index, r=self.r),
            add_counts=True
        )

    def flexible_radius_neighbors(self, radii:list, t:int):
        self.fg_index.add(self.repertoire)
        all_res = []
        for i in radii:
            print(f"Compute neighbors at TCRdist > {i}")
            res = tcr_dict_to_df(index_neighbors(query=self.repertoire, index=self.fg_index, r=i), add_counts=True)
            res = res.rename(columns={"neighbors":f"nn_{i}"})
            all_res.append(res)
            merged = reduce(
                lambda left,right: pd.merge(left,right,on=['v_call','junction_aa'], how='outer'),
                all_res
                )
        print(f"Compute argmin(d,T) > {t}")
        neighbors_at_radius = merged[merged.columns[2:]]
        merged["argmin_d"] = neighbors_at_radius.apply(lambda x: above_threshold(df=neighbors_at_radius,row=x,t=t), axis=1)
        merged = merged.dropna()
        merged["neighbors"] = [i[1][i[1].argmin_d] for i in merged.iterrows()]
        merged["argmin_d"] = merged["argmin_d"].apply(lambda x: float(x.split("_")[1]))
        self.nbr_counts = merged[["v_call","junction_aa","argmin_d","neighbors"]]
                

    def _setup_background_index(self, exhaustive=True, ratio=10):
        '''

        '''
        if self.bg_index is None:
            if self.background is None:
                depth = self.repertoire.shape[0] * ratio
                print(f'Background index not set up.\nSampling background of size {depth}.')
                # bg = match_vj_distribution(n=depth, foreground=self.repertoire)
                seq_gen = Background(repertoire=self.repertoire,factor=ratio,organism=self.organism)
                self.background = seq_gen.shuffled_background(num_workers=self.num_workers)
                print("Background contructed.")
            else:
                pass
            if exhaustive:
                self.bg_index = FlatIndex(encoder=self.encoder)
            else:
                depth = self.background.shape[0]
                k = int(depth/500)
                n = int(k/10)
                self.bg_index = IvfIndex(encoder=self.encoder, n_centroids=k, n_probe=n)
            self.bg_index.add(self.background)
            del self.background
        else:
            print(f'Using background of size {len(self.bg_index.ids)}.')

    def _prefilter(self, k):
        '''
        Filter out all foreground TCRs whose neighbor count is less than
        or equal to the background neighbor count.
        '''
        # Prepare prefilter index
        prefilter_index = FlatIndex(encoder=self.encoder)
        seq_gen = Background(repertoire=self.repertoire,factor=1)
        bg = seq_gen.shuffled_background()
        # bg = match_vj_distribution(n=self.rsize, foreground=self.repertoire)
        prefilter_index.add(bg)
        del bg
        # Compute neighbors
        seq_with_nbrs = self.nbr_counts[self.nbr_counts.neighbors>0]
        indexed = index_neighbors(query=seq_with_nbrs, index=prefilter_index, r=self.r)
        indexed = tcr_dict_to_df(indexed)

        # Prep results
        merged = seq_with_nbrs.merge(nbrs_in_background, on=['v_call', 'junction_aa'])
        filtered = merged[merged['neighbors_x']>merged['neighbors_y']]
        filtered = filtered.drop(columns=['neighbors_y']).rename(columns={'neighbors_x':'neighbors'})
        return filtered

    def _hypergeometric(self, data):
        '''
        Compute p-values for based on foreground and background neighbor counts
        using the hypergeometric distribution.

        NOTE:
        SciPy's hypergeometric sf function uses the boost distribution implementation.
        Integer values in the boost distribution use the C unsigned integer type, which for most
        compilers is 32 bits. Any values exceedingn 32 bits will prompt either nan or strange
        p-values.
        '''
        # Hypergeometric testing
        N = self.rsize - 1
        M = len(self.bg_index.ids) + N
        if M >= 2**32:
            import warnings
            warnings.warn(f"Population size exceeds {2**32-1} (32 bits).\n p-values may be miscalculated.")
        data['pval'] = data.apply(
            lambda x: hypergeom.sf(
                x['neighbors']-1, 
                M, 
                x['neighbors'] + x['background_neighbors'] + self.pseudocount, 
                N
                ),
            axis=1
            )
        return data.sort_values(by='pval')

    def foreground_neighbors_to_json(self, file):
        '''
        Dump foreground neighbor dictionary to json file.
        '''
        json_string = json.dumps(self.nbr_counts)
        f = open(file,"w")
        f.write(json_string)
        f.close()

    def foreground_neighbors_from_json(self, file):
        '''
        Load foreground neighbor dictionary from json file.
        '''
        with open(file) as json_file:
            self.nbr_counts = json.load(json_file)

    def save_background_index_to_file(self, name):
        '''
        Write the contents of an index to disk. This function
        will create a new folder containing one binary file
        that stores the faiss index, and one file that stores
        the TCR sequence ids.

        Parameters
        ----------
        name: str
            Name of the folder to save the files.
        '''
        assert self.bg_index is not None, "No background index found."
        os.mkdir(name)
        json_string = json.dumps(self.bg_index.ids)
        faiss.write_index(self.bg_index.idx, os.path.join(name,'index.bin'))
        f = open(os.path.join(name,'ids.json'),"w")
        f.write(json_string)
        f.close()

    def load_background_index_from_file(self, name:str, exhaustive:bool=True):
        '''
        Load a background index that is saved on disk.

        Parameters
        ----------
        name: str
            Name of the folder containing the files necessary for
            setting up an index (ids and faiss index).
        exhaustive: bool
            When True, use the exact IndexFlatL2, otherwise
            resort to the approximate IndevIVFFlat.
        '''
        faiss_index = faiss.read_index(os.path.join(name,"index.bin"))
        with open(os.path.join(name,"ids.json")) as json_file:
            index_ids = json.load(json_file)
        if exhaustive:
            self.bg_index = FlatIndex(encoder=self.encoder)
            self.bg_index.idx = faiss_index
            self.bg_index.ids = index_ids
        else:
            k = int(self.repertoire)/10
            n = int(k/10)
            self.bg_index = IvfIndex(encoder=self.encoder, n_centroids=k, n_probe=n)
            self.bg_index.idx = faiss_index
            self.bg_index.ids = index_ids

    def compute_pvalues(self, prefilter=False, q=1, ratio=10, fdr=1, exhaustive=False):
        '''
        Assign p-values to neighbor counts by contrasting against a background distribution.
        Uses a prefilter step (if True) to limit the number of vectors to query on the index.

        Parameters
        ----------
        prefilter: bool
            If True, uses a prefiltering step that eliminates TCRs whose foreground neighbor count
            is less than or equal to that of a size-matched background repertoire.
        depth
            Size of the background repertoire.
        fdr
            False discovery rate threshold.
        '''
        # Neighbor counts in foreground should be determined
        assert self.nbr_counts is not None, 'Please compute foreground neighbors first.'
        depth = self.rsize * ratio
        
        # Prefilter
        if prefilter:
            print('Prefiltering repertoire.')
            filtered = self._prefilter(k=int(depth/1000))
        else:
            pass
        filtered = self.nbr_counts[self.nbr_counts.neighbors>0]
        
        # Background
        print("Setting up background index...")
        self._setup_background_index(exhaustive=exhaustive, ratio=ratio)

        # Find neighbors in background
        print(f'Computing neighbor counts in background for {len(filtered)} sequences.')
        if "argmin_d" in filtered.columns:
            neighbor_dist = {}
            for i in filtered.argmin_d.unique():
                sub = filtered[filtered.argmin_d==i]
                neighbor_dist[i] = index_neighbors(query=sub, index=self.bg_index, r=i)
            comb = []
            for i in neighbor_dist:
                res_df = tcr_dict_to_df(neighbor_dist[i], add_counts=True)
                res_df["argmin_d"] = i
                res_df = res_df.rename(columns={"neighbors":"background_neighbors"})
                comb.append(res_df)
            comb = pd.concat(comb)
            filtered = filtered.merge(comb)
        else:
            nbrs_in_background = index_neighbors(query=filtered, index=self.bg_index, r=self.r)
            nbrs_in_background = tcr_dict_to_df(nbrs_in_background, add_counts=True)
            nbrs_in_background["neighbors"] = nbrs_in_background["neighbors"] * q
            nbrs_in_background = nbrs_in_background.rename(columns={"neighbors":"background_neighbors"})
            filtered = filtered.merge(nbrs_in_background)
        # Hypergeometric testing
        print('Performing hypergeometric testing.')
        p_values = self._hypergeometric(data=filtered)
        self.significant = p_values[p_values.pval<=fdr]
        # self.significant["background_neighbors"] = self.significant["background_neighbors"] / self.ratio
        # self.significant.columns = ["v_call","junction_aa","observed","expected","pval"]
        return self.significant

def to_dict(query, index):
    ndict = {}
    for i in range(len(query)):
        name = query.iloc[i].name
        q = pd.DataFrame(query.iloc[i]).T
        neighbors = index.radius_search(query=q, r=12.5)
        v = q.v_call.values[0]
        cdr3 = q.junction_aa.values[0]
        ndict[name] = tuple((v, cdr3, [i[0].split("_") + [i[1]] for i in neighbors]))
    return ndict

def shared(neighbors_a, neighbors_b):
    ovl = []
    # a vs b
    for i in neighbors_a:
        vcall = neighbors_a[i][0]
        junction_aa = neighbors_a[i][1]
        for j in neighbors_b:
            for k in neighbors_b[j][2]:
                if vcall == k[0]:
                    if junction_aa == k[1]:
                        ovl.append((i,j))
    # b vs a (reverse)
    for i in neighbors_b:
        vcall = neighbors_b[i][0]
        junction_aa = neighbors_b[i][1]
        for j in neighbors_a:
            for k in neighbors_a[j][2]:
                if vcall == k[0]:
                    if junction_aa == k[1]:
                        ovl.append((i,j))
    return list({tuple(sorted(item)) for item in ovl})

def to_neighborhood(neigh):
    return [neigh[0] + "_" + neigh[1]] + [i[0] + "_" + i[1] for i in neigh[2]]

def overlapping(neighbors_a, neighbors_b):
    result = []
    for i in neighbors_a:
        neighborhood_a = to_neighborhood(neighbors_a[i])
        for j in neighbors_b:
            neighborhood_b = to_neighborhood(neighbors_b[j])
            ovl = set(neighborhood_a).intersection(set(neighborhood_b))
            if len(ovl) > 0:
                result.append(tuple([(i,j),ovl]))
    return result


import numpy as np
import pandas as pd

from . import basic
from .basic import *
from .amino_acids import amino_acids
from . import all_genes
from . import tcrdist_svg_basic # hacky code duplication
from . import tcr_distances
from . import util
from . import tcr_sampler ## for analyze_junction

junction_bars = True

font_family = MONOSPACE_FONT_FAMILY # in basic, from ../convert_svg_to_png.py

greek_alpha = 'A'
greek_beta  = 'B'

# these don't always seem to work
#greek_alpha = '&#x3b1;'
#greek_beta  = '&#x3b2;'

junction_bars_color = { 'V':  'silver',
                        'N1': 'red',
                        'N':  'red',
                        'N2': 'red',
                        'D':  'black',
                        'J':  'dimgray' }

gap_character = '-' ## different from some other places

xmargin = 10
ymargin = 10

default_vj_logo_width = 200
default_pwmplusgaps_width = 770
default_xpad = 30

default_pwm_height = 100
default_junction_bars_height = 35.0 if junction_bars else 0.0
default_ypad = 3

default_ab_glyphs_spacer = 15

#used for rescaling below
default_width = 2*default_vj_logo_width + 2*default_xpad + default_pwmplusgaps_width
default_height = default_pwm_height + default_ypad + default_junction_bars_height


boxpad = 2 #OK to keep this fixed independent of scaling??

## returns
def make_tcr_logo(
        upper_left, # [x0,y0] location of the upper left corner
        tcrs, # list of tuples with tcr info in specific order, see docstring below
        members, # list of integers, indices into tcrs
        all_dists, # matrix of distances between tcrs
        ab,# A or B
        organism,
        rep_colors, # dictionary of colors indexed by the V and J gene names in the tcrs list
        vj_logo_width,
        pwmplusgaps_width,
        xpad,
        pwm_height,
        junction_bars_height,
        ypad,
        show_full_cdr3
):
    """ Returns a list of commands for svg output via tcrdist_svg_basic.create_file

    width = vj_logo_width + xpad + pwmplusgaps_width + xpad + vj_logo_width
    height = pwm_height + ypad + junction_bars_height

    tcrs = [( va_rep, ja_rep, vb_rep, jb_rep, cdr3a, cdr3b, cdr3a_nucseq_src, cdr3b_nucseq_src ), ...]


    """
    cmds = []

    def trim_gene_name_for_logo(gene, vj, ab=ab, organism=organism):
        if vj=='J':
            return gene[4:]
        if ( '_gd' in organism and ab=='B' or # could be TRAV or TRDV
             '_ig' in organism and ab=='A' ): # could be IGKV or IGLV
            return gene[2]+gene[4:]
        return gene[4:]


    assert ab in ['A','B']
    assert len(all_dists) == len(tcrs) and len(all_dists[0]) == len(tcrs)

    if ab== 'A':
        v_index, j_index, cdr3_index, nucseq_src_index = 0, 1, 4, 6
        junction_bars_order = ['V','N','J']
        greek_letter = greek_alpha
    else:
        assert ab=='B'
        v_index, j_index, cdr3_index, nucseq_src_index = 2, 3, 5, 7
        junction_bars_order = ['V','N1','D','N2','J']
        greek_letter = greek_beta

    real_size = len(members)

    distl = []
    for m1 in members:
        avgdis = 0.
        for m2 in members:
            avgdis += all_dists[m1][m2]
        avgdis/=len(members)
        distl.append( ( avgdis, m1 ) )
    distl.sort()

    center = distl[0][1]
    #print('center_tcr:', len(tcrs), distl[0], tcrs[center][v_index], tcrs[center][cdr3_index] )

    ## count v,j gene reps
    v_count = {}
    j_count = {}
    for rep in [tcrs[x][v_index] for x in members ]: v_count[rep] = v_count.get(rep,0)+1
    for rep in [tcrs[x][j_index] for x in members ]: j_count[rep] = j_count.get(rep,0)+1

    center_cdr3 = tcrs[center][ cdr3_index ]
    if not show_full_cdr3: center_cdr3 = center_cdr3[3:-2]
    L = len(center_cdr3)

    pwm = {}
    junction_pwm = {}
    gap_count = {}
    for i in range(L):
        pwm[i] = dict(list(zip(amino_acids+[gap_character],[0]*21)))
        gap_count[i]=0
    for i in range(3*L):
        junction_pwm[i] = dict( list(zip( junction_bars_order+[gap_character],
                                     [0.]*(1+len(junction_bars_order)))))

    for member in members:
        member_cdr3 = tcrs[member][cdr3_index]
        member_junction = tcrs[member][nucseq_src_index] ## a list
        if not show_full_cdr3:
            member_cdr3 = member_cdr3[3:-2]
            member_junction = member_junction[9:-6]
        assert len(member_junction) == 3*len(member_cdr3)
        a,b = tcr_distances.align_cdr3_regions( center_cdr3, member_cdr3, gap_character )

        for i in range(len(a)):
            if a[i] == gap_character:
                if i and a[i-1]==gap_character:continue
                gap_count[i-1] += 1

            else: ## b[i] could be gap_character or an amino_acid
                pwmpos = i - a[:i].count(gap_character)
                pwm[pwmpos][ b[i] ] += 1
                for j in range(3*pwmpos,3*pwmpos+3):
                    if b[i] == gap_character:
                        junction_pwm[j]['-'] += 1
                    else:
                        bpos = i-b[:i].count(gap_character)
                        assert b[i] == member_cdr3[bpos] ## sanity
                        bsrc = member_junction[3*bpos + j-3*pwmpos ]
                        junction_pwm[j][bsrc] += 1

    ## normalize the pwms
    for i in range(L):
        tot = float(sum(pwm[i].values()))
        for aa in pwm[i]:
            pwm[i][aa] /= tot
    for i in range(3*L):
        tot = float(sum(junction_pwm[i].values()))
        for aa in junction_pwm[i]:
            junction_pwm[i][aa] /= tot

    ## now we want to make a pwm
    total_gaps = sum(gap_count.values())


    column_width = pwmplusgaps_width / ( L + float(total_gaps)/len(members) )

    y0 = upper_left[1]
    y1 = y0 + pwm_height

    ## make a v-gene logo
    vl = [(y, trim_gene_name_for_logo(x, 'V'), rep_colors[x]) for x,y in v_count.items()]
    jl = [(y, trim_gene_name_for_logo(x, 'J'), rep_colors[x]) for x,y in j_count.items()]

    #single_glyph_width = 2* vj_logo_width + pwmplusgaps_width + 2*xpad
    #x0 = xmargin + glyph_size_text_width + ii_ab2 * ( single_glyph_width + ab_glyphs_spacer )
    x0 = upper_left[0]

    cmds.append( tcrdist_svg_basic.make_stack( (x0,y0), (x0 + vj_logo_width,y1), vl ) )

    ## a box around the V-logo
    cmds.append( tcrdist_svg_basic.rectangle( ( x0-boxpad,y0-boxpad), (x0 + vj_logo_width+boxpad,y1+boxpad ),
                                      'none', 'black', stroke_width=1 ) )

    if junction_bars: ## label V-logo down below
        text = 'V'+greek_letter
        fontsize = junction_bars_height*0.9
        p0 = [ x0 + 0.5* vj_logo_width - 0.6*fontsize, y1+junction_bars_height ]
        cmds.append( tcrdist_svg_basic.make_text( text, p0, fontsize, font_family=font_family ) )

    x0 += vj_logo_width + xpad

    cmds.append( tcrdist_svg_basic.rectangle( ( x0-boxpad,y0-boxpad), (x0 + pwmplusgaps_width+boxpad,y1+boxpad ),
                                      'none', 'black', stroke_width=1 ) )



    #print ic,L,size,y0,y1

    ## now show each column
    prev_gap_column_width = 0.0
    jb_rights = [] #debugging
    for pos in range(L):
        ## first the column of aas
        colpwm={}
        colpwm[0] = pwm[pos]
        #if verbose:
        #    print 'colpwm:',pos,pwm[pos]
        cmds.append( tcrdist_svg_basic.protein_logo( (x0,y0), (x0+column_width,y1), colpwm ) )

        save_x0 = x0 ## for junction_bars

        x0 += column_width

        ## any gaps?
        if gap_count[pos]:
            gap_column_width = float( column_width * gap_count[pos] ) / len(members)
            cmds.append( tcrdist_svg_basic.text_in_box( (x0,y0), (x0+gap_column_width,y1), gap_character, 'black' ) )

            x0+= gap_column_width
        else:
            gap_column_width = 0.0


        if junction_bars:
            junction_bar_width = ( column_width + gap_column_width/2. + prev_gap_column_width/2. )/3.
            junction_bar_x0 = save_x0 - prev_gap_column_width/2.
            # print 'left:',junction_bar_x0,'right:',junction_bar_x0 + 3.*junction_bar_width,\
            #     'prev_gap_column_width:',prev_gap_column_width,'gap_column_width:',gap_column_width,\
            #     'save_x0:',save_x0,'column_width:',column_width,'junction_bar_width:',junction_bar_width
            if jb_rights:
                assert abs( junction_bar_x0 - jb_rights[-1] )<1e-3
            jb_rights.append( junction_bar_x0 + 3.*junction_bar_width )

            y2 = y1+junction_bars_height
            for j in range(3):
                col = junction_pwm[3*pos+j]
                lcol = [ ( col[x],x) for x in junction_bars_order ]
                # lcol = [ (y,x) for x,y in col.iteritems()]
                # lcol.sort()
                # lcol.reverse()
                y1shift = y1+ ypad
                ## largest at the top
                for frac,a in lcol:
                    if a==gap_character: continue
                    y1shift_next = y1shift + frac * junction_bars_height
                    color = junction_bars_color[ a ]
                    p0 = [ junction_bar_x0+ j   *junction_bar_width, y1shift]
                    p1 = [ junction_bar_x0+(j+1)*junction_bar_width, y1shift_next ]
                    cmds.append( tcrdist_svg_basic.rectangle( p0, p1, fill=color, stroke=color ) )
                    y1shift = y1shift_next


        prev_gap_column_width = gap_column_width


    ## now the J-logo
    cmds.append( tcrdist_svg_basic.make_stack( (x0,y0), (x0+vj_logo_width,y1), jl ) )

    cmds.append( tcrdist_svg_basic.rectangle( ( x0-boxpad,y0-boxpad), (x0 + vj_logo_width+boxpad,y1+boxpad ),
                                      'none', 'black', stroke_width=1 ) )

    if junction_bars: ## label V-logo down below
        text = 'J'+greek_letter
        fontsize = junction_bars_height * 0.9
        p0 = [ x0 + 0.5* vj_logo_width - 0.6*fontsize, y1+junction_bars_height ]
        cmds.append( tcrdist_svg_basic.make_text( text, p0, fontsize, font_family=font_family ) )


    return cmds #####################################





def make_default_logo_svg_cmds(
        upper_left, width, height, organism, tcr_infos, chain,
        tcrdist_calculator = None,
        add_fake_alleles = False,
        show_full_cdr3 = False
):
    # right now single-chain only
    # returns cmds

    assert chain in 'AB'

    if tcrdist_calculator is None:
        tcrdist_calculator = tcr_distances.TcrDistCalculator(organism)

    util.assign_label_reps_and_colors_based_on_most_common_genes_in_repertoire( tcr_infos, organism )

    rep_colors = {}
    for info in tcr_infos:
        for vj in 'vj':
            for abl in 'ab':
                rep   = info[vj+abl+'_label_rep']
                color =  info[vj+abl+'_label_rep_color']
                rep_colors[rep] = color

    tcrs = []

    dist_tcrs = []

    def add_fake_allele_info(x):
        if '*' not in x:
            return x+'*01'
        else:
            return x

    for l in tcr_infos:

        mouse = None #l['subject']
        epitope = None #l['epitope']
        cdr3a = l['cdr3a']
        cdr3b = l['cdr3b']

        ## for computing distances
        va_gene = l['va_gene']
        ja_gene = l['ja_gene']
        va_genes = l['va_genes'].split(';')

        vb_gene = l['vb_gene']
        jb_gene = l['jb_gene']
        vb_genes = l['vb_genes'].split(';')


        # add '*01' -- hacky!
        if add_fake_alleles:
            va_genes = list(map( add_fake_allele_info, va_genes ))
            vb_genes = list(map( add_fake_allele_info, vb_genes ))
            va_gene = add_fake_allele_info( va_gene )
            ja_gene = add_fake_allele_info( ja_gene )
            vb_gene = add_fake_allele_info( vb_gene )
            jb_gene = add_fake_allele_info( jb_gene )

        va_reps = frozenset( ( all_genes.all_genes[organism][x].rep for x in va_genes ))
        vb_reps = frozenset( ( all_genes.all_genes[organism][x].rep for x in vb_genes ))

        if chain == 'A':
            dist_tcrs.append( (va_gene, ja_gene, cdr3a) )
        else:
            dist_tcrs.append( (vb_gene, jb_gene, cdr3b) )

        #dist_tcrs.append( [ va_reps, vb_reps, cdr3a, cdr3b ] )
        #all_info.append( l )


        ## note that we are using mm1 reps here that also dont have allele info
        va_rep = l['va_label_rep']
        ja_rep = l['ja_label_rep']
        vb_rep = l['vb_label_rep']
        jb_rep = l['jb_label_rep']

        cdr3a_nucseq_src = ['V']*(3*len(cdr3a)) ## hack, unused
        cdr3b_nucseq_src = ['V']*(3*len(cdr3b))
        if junction_bars:

            if chain == 'A':
                a_junction_results = tcr_sampler.analyze_junction( organism, va_gene, ja_gene,
                                                                   cdr3a, l['cdr3a_nucseq'].lower(),
                                                                   return_cdr3_nucseq_src=True )
                cdr3a_new_nucseq, cdr3a_protseq_masked, cdr3a_protseq_new_nucleotide_countstring,\
                    a_trims, a_inserts, cdr3a_nucseq_src = a_junction_results
            elif chain == 'B':
                b_junction_results = tcr_sampler.analyze_junction( organism, vb_gene, jb_gene,
                                                                   cdr3b, l['cdr3b_nucseq'].lower(),
                                                                   return_cdr3_nucseq_src=True )

                cdr3b_new_nucseq, cdr3b_protseq_masked, cdr3b_protseq_new_nucleotide_countstring,\
                    b_trims, b_inserts, cdr3b_nucseq_src = b_junction_results
                ## try to distinguish between N before D and N after D
                for i in range(len(cdr3b_nucseq_src)):
                    if cdr3b_nucseq_src[i] == 'N':
                        if cdr3b_nucseq_src[:i].count('D')==0:
                            cdr3b_nucseq_src[i] = 'N1'
                        else:
                            cdr3b_nucseq_src[i] = 'N2'


        assert len(cdr3a_nucseq_src) == 3*len(cdr3a)
        assert len(cdr3b_nucseq_src) == 3*len(cdr3b)
        #print cdr3b, cdr3b_nucseq_src
        tcrs.append( ( va_rep, ja_rep, vb_rep, jb_rep, cdr3a, cdr3b,
                       cdr3a_nucseq_src, cdr3b_nucseq_src ) )


    ## compute distances, used in logo construction for picking the center tcr for aligning against
    #print 'computing distances:',len(dist_tcrs)
    all_dists = np.zeros( ( len(dist_tcrs), len(dist_tcrs)) )
    for i,t1 in enumerate( dist_tcrs ):
        for j in range(i+1,len(dist_tcrs)):
            dist = tcrdist_calculator.single_chain_distance(t1, dist_tcrs[j])
            all_dists[i][j] = dist
            all_dists[j][i] = dist

    # now make the logo
    members = list(range(len(tcrs)))


    scale_w = float( width ) / default_width
    scale_h = float( height ) / default_height

    # scale everything by our desired height, width

    return make_tcr_logo( upper_left, tcrs, members, all_dists, chain, organism, rep_colors,
                          scale_w * default_vj_logo_width,
                          scale_w * default_pwmplusgaps_width,
                          scale_w * default_xpad,
                          scale_h * default_pwm_height,
                          scale_h * default_junction_bars_height,
                          scale_h * default_ypad,
                          show_full_cdr3 )




def make_tcr_logo_svg_commands_for_tcrs(
        tcrs,
        chain,
        organism,
        upper_left,
        width_height,
        tcrdist_calculator=None # so we can avoid recalculating V gene distances over and over again
):
    ''' tcrs is a list of tuples: tcrs = [ (atcr1,btcr1), (atcr2,btcr2), ....
    atcr1 = (va1, ja1, cdr3a1, cdr3a_nucseq1, *)
    btcr1 = (vb1, jb1, cdr3b1, cdr3b_nucseq1, *)

    upper_left and width_height are tuples (x,y)

    returns the svg cmds

    untested, currently under development
    '''
    assert chain in 'AB'

    add_fake_alleles = False
    show_full_cdr3 = False

    # SILLY backwards compatibility, fill out info that would have been in clones file
    infos = []
    for tcr in tcrs:
        atcr, btcr = tcr
        info = {}
        for iab, ab in enumerate('ab'):
            info[f'cdr3{ab}'] = tcr[iab][2]
            info[f'cdr3{ab}_nucseq'] = tcr[iab][3]
            for ivj, vj in enumerate('vj'):
                gene = tcr[iab][ivj]
                assert gene in all_genes.all_genes[organism]
                for tag in 'gene genes rep reps'.split():
                    info[f'{vj}{ab}_{tag}'] = gene
                info[f'{vj}{ab}_countreps'] = all_genes.all_genes[organism][gene].count_rep
        infos.append(info)

    if tcrdist_calculator is None:
        tcrdist_calculator = tcr_distances.TcrDistCalculator(organism)

    cmds = make_default_logo_svg_cmds(
        upper_left, width_height[0], width_height[1], organism, infos, chain,
        tcrdist_calculator = tcrdist_calculator,
        add_fake_alleles = add_fake_alleles, show_full_cdr3 = show_full_cdr3 )


    return cmds


def make_tcr_logo_for_tcrs(
        tcrs,
        chain,
        organism,
        pngfile,
        tcrdist_calculator=None # so we can avoid recalculating V gene distances over and over again
):
    ''' tcrs is a list of tuples: tcrs = [ (atcr1,btcr1), ....
    atcr1 = (va1, ja1, cdr3a1,*)
    btcr1 = (vb1, jb1, cdr3b1,*)
    '''
    assert chain in 'AB'
    assert pngfile.endswith('.png')

    add_fake_alleles = False
    show_full_cdr3 = False

    # backwards compatibility, fill out info that would have been in clones file
    infos = []
    for tcr in tcrs:
        atcr, btcr = tcr
        info = {}
        for iab, ab in enumerate('ab'):
            info[f'cdr3{ab}'] = tcr[iab][2]
            info[f'cdr3{ab}_nucseq'] = tcr[iab][3]
            for ivj, vj in enumerate('vj'):
                gene = tcr[iab][ivj]
                assert gene in all_genes.all_genes[organism]
                for tag in 'gene genes rep reps'.split():
                    info[f'{vj}{ab}_{tag}'] = gene
                info[f'{vj}{ab}_countreps'] = all_genes.all_genes[organism][gene].count_rep
        infos.append(info)

    if tcrdist_calculator is None:
        tcrdist_calculator = tcr_distances.TcrDistCalculator(organism)

    cmds = make_default_logo_svg_cmds(
        [xmargin,ymargin], default_width, default_height, organism, infos,
        chain,tcrdist_calculator = tcrdist_calculator,
        add_fake_alleles = add_fake_alleles, show_full_cdr3 = show_full_cdr3 )


    svg_width = 2*xmargin + default_width
    svg_height = 2*ymargin + default_height

    tcrdist_svg_basic.create_file( cmds, svg_width, svg_height, pngfile[:-4]+'.svg', create_png=True )

    #print default_width,default_height


########################################################################################################################
########################################################################################################################
########################################################################################################################
########################################################################################################################

if __name__ == '__main__':

    # with Parser(locals()) as p:
    #     #p.str('args').unspecified_default().multiple().required()
    #     p.str('clones_file').required()
    #     p.str('organism').required()
    #     p.str('outfile_prefix')
    #     p.flag('add_fake_alleles')       # --flag_arg  (no argument passed)
    #     p.flag('show_full_cdr3')       # --flag_arg  (no argument passed)
    #     #p.flag('verbose')       # --flag_arg  (no argument passed)
    #     p.multiword('ABs').cast(lambda x: x.split())
    #     p.multiword('epitopes').cast(lambda x: x.split())
    clones_file = 'tmpclones.tsv'
    organism = 'human'
    add_fake_alleles = False
    show_full_cdr3 = False
    AB = 'A'
    outfile_prefix = None

    junction_bars = True

    if outfile_prefix is None:
        outfile_prefix = clones_file[:-4]

    df = pd.read_csv( clones_file, sep='\t' )
    infos = [ x._asdict() for x in df.itertuples() ]

    for l in infos:
        for vj in 'vj':
            for ab in 'ab':
                countreps_tag = '{}{}_countreps'.format(vj, ab)
                if countreps_tag not in l:
                    genes_tag = '{}{}_genes'.format(vj, ab)
                    genes = l[genes_tag].split(';')
                    l[countreps_tag] = ';'.join(util.countreps_from_genes(genes, organism))

    tcrdist_calculator = tcr_distances.TcrDistCalculator(organism)

    cmds = make_default_logo_svg_cmds( [xmargin,ymargin], default_width, default_height, organism, infos, AB,
                                       tcrdist_calculator = tcrdist_calculator,
                                       add_fake_alleles = add_fake_alleles, show_full_cdr3 = show_full_cdr3 )


    svg_width = 2*xmargin + default_width
    svg_height = 2*ymargin + default_height

    tcrdist_svg_basic.create_file( cmds, svg_width, svg_height, outfile_prefix+'.svg', create_png=True )

    #print default_width,default_height



from abc import ABC
from functools import partial
from typing import List, Tuple, Callable, Union

import os
import faiss
# from pynndescent import NNDescent
import numpy as np
import pandas as pd

from .encoding import TCRDistEncoder
from .analysis import TcrCollection
from .constants.preprocessing import detect_vgene_col, detect_cdr3_col

from timeit import default_timer as timer

def convert_range_search_output(lims, D, I, offset=0):
    '''
    Convert the output of a range search to a list of tuples.
    '''
    # Calculate the number of queries
    num_queries = len(lims) - 1
    # Vectorize the construction of result pairs
    result_indices = []
    result_distances = []
    for i in range(num_queries):
        start = lims[i]
        end = lims[i + 1]
        query_indices = np.full(end - start, i)
        neighbor_indices = I[start:end] + offset
        distances = D[start:end]
        result_indices.extend(zip(query_indices, neighbor_indices))
        result_distances.extend(distances)
    result = list(zip(result_indices, result_distances))
    return result


class BaseIndex(ABC):
    """
    Abstract structure for an index, supports adding TCRs and searching them.
    This index structure allows searching in two ways:
        - kNN retrieval: returns k-nearest neighbors for any given query
        - radius search: returns all TCRs within a radius r of the query
    """

    def __init__(
            self, idx: faiss.Index, 
            encoder: TCRDistEncoder,
            ) -> None:
        super().__init__()
        self.idx = idx
        self.encoder = encoder
        self.ids = {}
        self.n = 0

    def _add_hashes(self, hashes):
        if not self.idx.is_trained:
            self.idx.train(hashes)
        self.idx.add(hashes)

    def _add_ids(self, X):
        if isinstance(self.encoder, TCRDistEncoder):
            if isinstance(X, pd.DataFrame):
                if self.encoder.full_tcr:
                    for _, x in X.iterrows():
                        if self.encoder.chain == 'AB':
                            self.ids[self.n] = x['va'] + "_" + x['cdr3a'] + "_" + x['vb'] + "_" + x['cdr3b']
                            self.n += 1                            
                        else:
                            if self.encoder.vgene_col is None:
                                vgene_col = detect_vgene_col(X)
                            else:
                                vgene_col = self.encoder.vgene_col
                            if self.encoder.cdr3_col is None:
                                cdr3_col = detect_cdr3_col(X)
                            else:
                                cdr3_col = self.encoder.cdr3_col
                            self.ids[self.n] = x[vgene_col] + "_" + x[cdr3_col]
                            self.n += 1
            else:
                for x in X:
                    self.ids[self.n] = x
                    self.n += 1
        else:
            for x in X:
                self.ids[self.n] = x
                self.n += 1

    def add(self, tcrs: TcrCollection, vecs:np.ndarray=None):
        """
        Add sequences to the index.

        Parameters
        ----------
        tcrs : TcrCollection
            Collection of TCRs to add. Can be a Repertoire, list of Clusters, or
            list of str.
        vecs : np.ndarray, optional
            Precomputed vectors for the TCRs. If not passed, the encoder will
            be used to compute the vectors.
        """
        if isinstance(tcrs, pd.DataFrame):
            if not self.encoder.full_tcr:
                import warnings
                warnings.warn(f"Provided DataFrame but 'full_tcr' was set to {self.encoder.full_tcr} --> only using CDR3 sequences to create embedding.")
        # If vecs are not passed, encode the TCRs
        if vecs is None:
            vecs = self.encoder.transform(tcrs).astype(np.float32)
        # Add the vecs and TCR ids to the index
        self._add_hashes(vecs)
        self._add_ids(tcrs)
        return self

    def _assert_trained(self):
        if not self.idx.is_trained:
            raise ValueError("Index is untrained, please add first.")

    def _search(self, x, k):
        return self.idx.search(x=x, k=k)

    def knn_search(self, y: TcrCollection, k: int = 100):
        """
        Search index for k-nearest neighbours.

        Parameters
        ----------
        y : TcrCollection
            Query TCRs.
        k : int, default = 100
            Number of nearest neighbours to search.

        Returns
        -------
        KnnResult
            `KnnResult` object.
        """
        self._assert_trained()
        hashes = self.encoder.transform(y).astype(np.float32)
        D, I = self._search(x=hashes, k=k)
        return KnnResult(y, D, I, self.ids)

    def _within_radius(self, x, r):
        if len(x.shape) == 1:
            xq = np.expand_dims(x, axis=0)
        else:
            xq = x
        return self.idx.range_search(x=xq, thresh=r)

    def _report_radius(self, I, D, exclude_self:bool=True):
        if exclude_self:
            return [(self.ids[i],d) for (i,d) in zip(I,D) if d > 0]
        else:
            return [(self.ids[i],d) for (i,d) in zip(I,D)]
        
    def search(self, vecs, r:float):
        lims, D, I = self._within_radius(x=vecs, r=r)
        return lims, D, I

    def radius_search(self, query:Union[str,list,pd.DataFrame], r:float, exclude_self=True):
        """
        ONLY COMPATIBLE WITH THE FOLLOWING INDEX STRUCTURES:
        - IndexFlat
        - IndexIVFFlat
        - IndexScalarQuantizer
        - IndexIVFScalarQuantizer
        Retrieve all items that are within a radius around the query point.

        Parameters
        ----------
        query
            TCR sequence(s) set as the centroid around which to query.
        r : float
            Radius.
        """
        q = self.encoder.transform(query).astype(np.float32)
        lims, D, I = self._within_radius(x=q, r=r)
        return self._report_radius(I, D, exclude_self=exclude_self)

    def array_search(self, query:pd.DataFrame, r:Union[float,int], exclude_self=False):
        xq = self.encoder.transform(query).astype(np.float32)
        lims, D, I = self.idx.range_search(x=xq, thresh=r)
        result = []
        n = 0
        for i,j in enumerate(query.index):
            nn = lims[i+1]-lims[i]
            nbr_ids = I[int(n):int(n+nn)]
            dist = D[int(n):int(n+nn)]
            n += nn
            clone_id = '_'.join(query.loc[j][['v_call', 'junction_aa']])
            nbrs = [self.ids[j] for j in nbr_ids]
            # result[clone_id] = (nbrs,dist)
            
            nbr_df = pd.DataFrame({'target':nbrs,'distance':dist}).sort_values(by='distance')
            nbr_df['source'] = clone_id
            result.append(nbr_df)

        col_order = ['source', 'target', 'distance']
        result = pd.concat(result)
        result = result[col_order]
        # Remove any self-sequences that are identical to the query sequence
        if exclude_self:
            return result[result.distance > 0]
        return result

    def save(self, filepath):
        '''
        Write the contents of an index to disk. This function
        will create a new folder containing one binary file
        that stores the faiss index, and one file that stores
        the TCR sequence ids.

        Parameters
        ----------
        name: str
            Name of the folder to save the files.
        '''
        assert self.n > 0, "Index is empty."
        faiss.write_index(self.idx, os.path.join(filepath))

    def load(self, filepath):
        '''
        Load an index from disk.
        '''
        self.idx = faiss.read_index(filepath)


class FlatIndex(BaseIndex):
    """
    Exact search for euclidean hash disstrtance.
    """

    def __init__(self, encoder: TCRDistEncoder) -> None:
        """
        Initialize index.

        Parameters
        ----------
        encoder : TCRDistEncoder
            Fitted TCRDistEncoder class.
        """
        idx = faiss.IndexFlatL2(encoder.m)
        super().__init__(idx, encoder)


class PynndescentIndex(BaseIndex):
    """
    Approximate search using PyNNDescent.
    """

    def __init__(
        self,
        encoder: TCRDistEncoder,
        k: int = 100,
        diversify_prob: float = 1.0,
        pruning_degree_multiplier: float = 1.5,
    ):
        """
        Initialize index.
        """
        idx = partial(
            NNDescent,
            n_neighbors=k,
            diversify_prob=diversify_prob,
            pruning_degree_multiplier=pruning_degree_multiplier,
        )
        idx.is_trained = True
        super().__init__(idx, encoder)


    def _add_hashes(self, X):
        self.idx = self.idx(X)
        self.idx.is_trained = True
    
    def _search(self, y, k):
        I, D = self.idx.query(y, k=k)
        return D,I

    def _search_self(self):
        I, D = self.idx.neighbor_graph
        return KnnResult(self.ids.values(), D, I, self.ids)


    def knn_search(self, y: TcrCollection=None):
        """
        Search index for nearest neighbours.

        Parameters
        ----------
        y : TcrCollection, optional
            The query TCRs. If not passed, returns the neighbours within the
            data added to the index, which is much faster.
        """
        if not y:
            return self._search_self()
        return super().knn_search(y)
    

class BaseApproximateIndex(BaseIndex):
    """
    Abstract class for approximate indexes implementing the `n_probe` property.
    """

    @property
    def n_probe(self):
        return faiss.extract_index_ivf(self.idx).nprobe

    @n_probe.setter
    def n_probe(self, n: int):
        ivf = faiss.extract_index_ivf(self.idx)
        ivf.nprobe = n 


class IvfIndex(BaseApproximateIndex):
    def __init__(
        self, encoder: TCRDistEncoder, n_centroids: int = 32, n_probe: int = 5
    ) -> None:
        """
        Inverted file index for approximate nearest neighbour search.

        Parameters
        ----------
        encoder : TCRDistEncoder
            Fitted encoder object to transform CDR3 to vectors.
        n_centroids : int, default=32
            Number of centroids for the initial k-means clustering.
        n_probe : int, default=5
            Number of centroids to search at query time. Higher n_probe means
            higher recall, but slower speed.
        """
        idx = faiss.index_factory(encoder.m, f"IVF{n_centroids},Flat")
        super().__init__(idx, encoder)
        self.n_probe = n_probe


class HnswIndex:
    def __init__(self, encoder: TCRDistEncoder, n_links: int = 32) -> None:
        """
        Index based on Hierarchical Navigable Small World networks.

        Parameters
        ----------
        encoder : TCRDistEncoder
            Fitted encoder object to transform CDR3 to vectors.
        n_links : int, default=32
            Number of bi-directional links created for each element during index
            construction. Increasing M leads to better recall but higher memory
            size and slightly slower searching.

        """
        idx = faiss.index_factory(64, f"HNSW{n_links},Flat")
        super().__init__(idx, encoder)


class FastApproximateIndex(BaseApproximateIndex):
    def __init__(
        self,
        encoder: TCRDistEncoder,
        n_centroids: int = 256,
        n_links: int = 32,
        n_probe: int = 10,
    ) -> None:
        """
        Approximate index based on a combination of IVF and HNSW using scalar
        quantizer encoding.

        Parameters
        ----------
        encoder : TCRDistEncoder
            Fitted encoder object to transform CDR3 to vectors.
        n_centroids : int, default=32
            Number of centroids for the initial k-means clustering.
        n_probe : int, default=5
            Number of centroids to search at query time. Higher n_probe means
            higher recall, but slower speed.
        n_links : int, default=32
            Number of bi-directional links created for each element during index
            construction. Increasing M leads to better recall but higher memory
            size and slightly slower searching.
        """
        idx = faiss.index_factory(64, f"IVF{n_centroids}_HNSW{n_links},SQ6")
        super().__init__(idx, encoder)
        self.n_probe = n_probe


class KnnResult:
    """
    Result of k-nearest neighbor search.
    """

    def __init__(self, y, D, I, ids) -> None:
        self.D = np.sqrt(D)
        self.I = I
        self.y_idx = {y: i for i, y in enumerate(y)}
        self.ids = ids
        query_size, k = D.shape
        self.query_size = query_size
        self.k = k

    def __repr__(self) -> str:
        return f"k-nearest neighbours result (size={self.query_size}, k={self.k})"

    def _extract_neighbours(self, cdr3:str):
        try:
            i = self.y_idx[cdr3]
        except KeyError:
            raise KeyError(f"{cdr3} was not part of your query")
        I_ = np.vectorize(self._annotate_id)(self.I[i])
        return i, I_

    def extract_neighbours(self, cdr3: str) -> List[Tuple[str, float]]:
        """
        Query the KnnResult for neighbours of a specific sequence.

        Parameters:
        -----------
        cdr3 : str
            Query sequence.

        Returns
        -------
        List[(str, float)]
            List of matches, containing (sequence, score) tuples.
        """
        i, I_ = self._extract_neighbours(cdr3)
        return list(zip(I_, self.D[i]))

    def extract_neighbour_sequences(self, cdr3:str) -> List[str]:
        return self._extract_neighbours(cdr3)[1]

    def _annotate_id(self, cdr3_id):
        return self.ids.get(cdr3_id)

    def _refine_edges_iterator(self, distance_function, threshold):
        for s1 in self.y_idx.keys():
            seqs = self.extract_neighbour_sequences(s1)
            for match in ((s1, s2, dist) for s2 in seqs if (dist := distance_function(s1, s2)) <= threshold):
                yield match

    def _iter_edges(self):
        pass

    def refine(self, distance_function:Callable, threshold:float, k:int=None) -> pd.DataFrame:
        """
        Perform a second round refinement of k-nearest neighbour matches, using a custom distance function and threshold.

        Parameters:
        distance_function : Callable,
            A function taking in two string arguments, returning the distance between the sequences.
        threshold : float
            Only sequence pairs at or below this distance are retained.
        k : int, optional
            Only the k closest matches for each query sequence are retained, if below the threshold.
        """
        df = pd.DataFrame(self._refine_edges_iterator(distance_function, threshold))
        if not df.empty:
            df.columns = ["query_cdr3", "match_cdr3", "distance"]
        if k:
            df = df.sort_values('distance').groupby("query_cdr3", sort=False).head(k).sort_index()
        return df

    def as_network(self, max_edge: float = 15):
        raise NotImplementedError()



import pandas as pd

from .analysis import Repertoire


def read_AIRR(
    filepath: str,
    filter_productive: bool = True,
    filter_TRB: bool = True,
    filter_min_duplicate_count: int = 0,
) -> Repertoire:
    """
    Read in data from an AIRR formatted `.tsv` file.

    Parameters
    ----------
    filepath : str
        Path of file.
    filter_productive : bool, default = True
        Retain only productive sequences.
    filter_TRB : bool, default = True
        Retain only TCR beta chains, inferred from the V and J-gene calls.
    filter_min_duplicate_count : int, optional
        Retain only sequences observed more than n times.


    Returns
    -------
    Repertoire
        Repertoire object containing the data.
    """

    cols = [
        "sequence_id",
        "productive",
        "v_call",
        "j_call",
        "duplicate_count",
        "junction_aa",
    ]
    df = pd.read_csv(filepath, sep="\t", usecols=cols)
    df = df.set_index("sequence_id")

    if "productive" in df:
        if df["productive"].dtype == "O":
            df["productive"] = df["productive"] == "T"

    if filter_productive:
        df = df.query("productive == True")

    if filter_min_duplicate_count:
        df = df.query(f"duplicate_count > {filter_min_duplicate_count}")

    df = df.drop("productive", axis=1)

    if filter_TRB:
        df = df.query('v_call.str.contains("TRB") or j_call.str.contains("TRB")')

    return Repertoire(df)


def read_OLGA(filepath: str) -> Repertoire:
    """
    Read in data from an OLGA-generated `.tsv` file.

    Parameters
    ----------
    filepath : str
        Path of file.

    Returns
    -------
    Repertoire
        Repertoire object containing the data.
    """
    df = pd.read_csv(filepath, sep="\t", header=None)
    df = df.drop(0, axis=1)
    df.columns = ["junction_aa", "v_call", "j_call"]
    return Repertoire(df)

def read_vdjdb(filepath:str, filter_TRB:bool = True, filter_human : bool = True, exclude_10x:bool=False, exclude_studies:list=[]) -> Repertoire:
    """
    Read the vdjdb.slim.txt file into a Repertoire object

    Parameters
    ----------
    filepath : str
        Location of the vdjdb database file.
    filter_TRB : bool, default = True
        Retain only TRB sequences.
    filter_human : bool, default = True
        Retain only human-derived sequences.
    exclude_10x : bool, optional
        Exclude sequence annotations derived only from the 10X genomics "A new way of exploring immunity" study.
    exclude_studies : list, optional
        List containing reference_ids of other studies to exclude.
    """
    df =  pd.read_csv(filepath, sep="\t")
    df.columns = df.columns.str.replace(".", "_", regex=False)

    if exclude_10x:
        exclude_studies.append('https://www.10xgenomics.com/resources/application-notes/a-new-way-of-exploring-immunity-linking-highly-multiplexed-antigen-recognition-to-immune-repertoire-and-phenotype/#')

    query_filter = '(reference_id not in @exclude_studies)'
    if filter_TRB:
        query_filter += ' and (gene == "TRB")'
    if filter_human:
        query_filter += ' and (species == "HomoSapiens")'

    df = (
        df
        .query(query_filter)
        .rename({"v_segm":"v_call", "j_segm":"j_call", "cdr3":"junction_aa"}, axis="columns")
        )

    return Repertoire(df)


def read_clustcr(filepath: str) -> Repertoire:
    raise NotImplementedError()



## http://python.genedrift.org/2007/04/19/genetic-code-part-i/

gencode = {
    'ATA':'I',    #Isoleucine
    'ATC':'I',    #Isoleucine
    'ATT':'I',    # Isoleucine
    'ATG':'M',    # Methionine
    'ACA':'T',    # Threonine
    'ACC':'T',    # Threonine
    'ACG':'T',    # Threonine
    'ACT':'T',    # Threonine
    'AAC':'N',    # Asparagine
    'AAT':'N',    # Asparagine
    'AAA':'K',    # Lysine
    'AAG':'K',    # Lysine
    'AGC':'S',    # Serine
    'AGT':'S',    # Serine
    'AGA':'R',    # Arginine
    'AGG':'R',    # Arginine
    'CTA':'L',    # Leucine
    'CTC':'L',    # Leucine
    'CTG':'L',    # Leucine
    'CTT':'L',    # Leucine
    'CCA':'P',    # Proline
    'CCC':'P',    # Proline
    'CCG':'P',    # Proline
    'CCT':'P',    # Proline
    'CAC':'H',    # Histidine
    'CAT':'H',    # Histidine
    'CAA':'Q',    # Glutamine
    'CAG':'Q',    # Glutamine
    'CGA':'R',    # Arginine
    'CGC':'R',    # Arginine
    'CGG':'R',    # Arginine
    'CGT':'R',    # Arginine
    'GTA':'V',    # Valine
    'GTC':'V',    # Valine
    'GTG':'V',    # Valine
    'GTT':'V',    # Valine
    'GCA':'A',    # Alanine
    'GCC':'A',    # Alanine
    'GCG':'A',    # Alanine
    'GCT':'A',    # Alanine
    'GAC':'D',    # Aspartic Acid
    'GAT':'D',    # Aspartic Acid
    'GAA':'E',    # Glutamic Acid
    'GAG':'E',    # Glutamic Acid
    'GGA':'G',    # Glycine
    'GGC':'G',    # Glycine
    'GGG':'G',    # Glycine
    'GGT':'G',    # Glycine
    'TCA':'S',    # Serine
    'TCC':'S',    # Serine
    'TCG':'S',    # Serine
    'TCT':'S',    # Serine
    'TTC':'F',    # Phenylalanine
    'TTT':'F',    # Phenylalanine
    'TTA':'L',    # Leucine
    'TTG':'L',    # Leucine
    'TAC':'Y',    # Tyrosine
    'TAT':'Y',    # Tyrosine
    'TAA':'_',    # Stop
    'TAG':'_',    # Stop
    'TGC':'C',    # Cysteine
    'TGT':'C',    # Cysteine
    'TGA':'_',    # Stop
    'TGG':'W',    # Tryptophan
    }



## http://www.pasteur.fr/formation/infobio/python/ch15.html

code = {'ttt': 'F', 'tct': 'S', 'tat': 'Y', 'tgt': 'C',
        'ttc': 'F', 'tcc': 'S', 'tac': 'Y', 'tgc': 'C',
        'tta': 'L', 'tca': 'S', 'taa': '*', 'tga': '*',
        'ttg': 'L', 'tcg': 'S', 'tag': '*', 'tgg': 'W',
        'ctt': 'L', 'cct': 'P', 'cat': 'H', 'cgt': 'R',
        'ctc': 'L', 'ccc': 'P', 'cac': 'H', 'cgc': 'R',
        'cta': 'L', 'cca': 'P', 'caa': 'Q', 'cga': 'R',
        'ctg': 'L', 'ccg': 'P', 'cag': 'Q', 'cgg': 'R',
        'att': 'I', 'act': 'T', 'aat': 'N', 'agt': 'S',
        'atc': 'I', 'acc': 'T', 'aac': 'N', 'agc': 'S',
        'ata': 'I', 'aca': 'T', 'aaa': 'K', 'aga': 'R',
        'atg': 'M', 'acg': 'T', 'aag': 'K', 'agg': 'R',
        'gtt': 'V', 'gct': 'A', 'gat': 'D', 'ggt': 'G',
        'gtc': 'V', 'gcc': 'A', 'gac': 'D', 'ggc': 'G',
        'gta': 'V', 'gca': 'A', 'gaa': 'E', 'gga': 'G',
        'gtg': 'V', 'gcg': 'A', 'gag': 'E', 'ggg': 'G'
        }

genetic_code = {}
reverse_genetic_code = {}

for codon in gencode:
    lowcodon = codon.lower()
    assert code[ lowcodon ] == gencode[ codon ] or ( gencode[ codon ] == '_' and code[ lowcodon ] == '*' )
    aa = code[ lowcodon ]
    genetic_code[ lowcodon ] = aa
    if aa not in reverse_genetic_code: reverse_genetic_code[aa] = []
    reverse_genetic_code[aa].append( lowcodon )

if 1:
    ## this is partial, just what's needed for the code below...
    degmap = {
        'a':'a',
        'c':'c',
        'g':'g',
        't':'t',
        'ct':'y',
        'ag':'r',
        'act':'h',
        'acgt':'n'
    }


    aa2degenerate_codons = {}
    for aa,codons in reverse_genetic_code.items():
        if aa=='*':continue
        ## write this as one or more degenerate strings
        degenerate_codons = []

        xys = set( [ x[:2] for x in codons] )

        for xy in xys:
            thirdnucs = ''.join(
                sorted(set((x[2] for x in codons if x[:2] == xy ))))
            degenerate_codons.append( xy + degmap[ thirdnucs ] )

        #print(aa, degenerate_codons)
        aa2degenerate_codons[aa] = degenerate_codons




"""
Test export result of vectorized TCRdist to sparse matrix
with dimensions matching input forground sequences. 

conda install pytest
conda install pytest-cov
pytest -v raptcr/tests/test_export.py
"""
import pandas as pd
import numpy as np
import os
import time
from scipy.sparse import csr_matrix

from raptcr.neighbors import NeighborEnrichment
from raptcr.export import index_neighbors_manual
from raptcr.export import range_search_to_csr_matrix

def test_export_to_csr_matrix():
    """
    integration test of (1) raptcr.export.index_neighbors_manual 
    with (2) raptcr.export.range_search_to_csr_matrix
    """
    start_time = time.perf_counter()
    foreground = pd.read_table('raptcr/datasets/1K_sequences.tsv')
    foreground['v_call'] = foreground['v_call'].apply(lambda x : f"{x}*01")
    foreground['j_call'] = foreground['j_call'].apply(lambda x : f"{x}*01")
    enricher = NeighborEnrichment(repertoire=foreground)
    enricher.fixed_radius_neighbors(radius=37) # Determine neighbors in foreground
    end_time = time.perf_counter()
    elapsed_time = end_time - start_time
    print(f"Elapsed time: {elapsed_time} seconds")
    assert(enricher.repertoire.shape[0] == foreground.shape[0])
    assert((foreground.junction_aa ==enricher.repertoire.junction_aa).all())
    lims, D, I  = index_neighbors_manual(query= enricher.repertoire, index=enricher.fg_index, r= enricher.r)
    csr_mat = range_search_to_csr_matrix(lims, D, I)
    assert(isinstance(csr_mat, csr_matrix))




import os
from os.path import exists
import pandas as pd
#from . import basic
#from .amino_acids import amino_acids
from .all_genes import all_genes
from .genetic_code import aa2degenerate_codons
from . import tcr_sampler
from . import logo_tools
from . import translation

# need to choose a nucleotide in N insertions
# could get fancier: use nucleotide context, whether v-d d-j or v-j
#  (which determines strand)
#
N_insertion_frequency_order = {'c':0, 'g':1, 't':2, 'a':3}

def get_degenerate_cdr3_nucseqs(cdr3):
    ''' Convert from amino acid cdr3 to list of degenerate coding seqs
    '''

    seqs= ['']
    while cdr3:
        seqs = [x+y for x in seqs for y in aa2degenerate_codons[cdr3[0]]]
        cdr3 = cdr3[1:]

    return seqs

def infer_cdr3_nucleotides(
        organism,
        v,
        j,
        cdr3,
        verbose=False,
        return_stats=False,
        allow_internal_mismatches=False,
        allow_allele_changes=False,
):
    ''' returns cdr3_nucseq and (optionally) the number of insertions

    if allow_allele_changes: returns cdr3_nucseq, jgene, and (optionally) inserts

    Still need to IMPROVE the logic to fixup alleles based on protein seq
    matches?
    '''

    ########################### special handling of common allelic aa variation ########
    if allow_allele_changes: # need to return the j gene
        special_human_jgenes = ['TRAJ24','TRAJ37']
        jg = j.split('*')[0]
        if organism == 'human' and jg in special_human_jgenes:
            best_inserts = 1000
            for allele in ['*01','*02']:
                new_j = jg+allele
                for aim in [True, False]:
                    cdr3_nucseq, total_inserts = infer_cdr3_nucleotides(
                        organism, v, new_j, cdr3, return_stats=True,
                        allow_internal_mismatches=aim,
                        allow_allele_changes=False,
                    )
                    if allow_internal_mismatches:
                        total_inserts += 1
                    #print(j, allow_internal_mismatches, best_inserts, total_inserts,
                    #      cdr3, cdr3_nucseq)
                    if total_inserts < best_inserts:
                        best_inserts = total_inserts
                        new_tcr = (v, new_j, cdr3, cdr3_nucseq)
            _, new_j, _, cdr3_nucseq = new_tcr
        else:
            new_j = j
            cdr3_nucseq, best_inserts = infer_cdr3_nucleotides(
                organism, v, new_j, cdr3, return_stats=True,
                allow_internal_mismatches=allow_internal_mismatches,
                allow_allele_changes=False,
            )

        if return_stats:
            return cdr3_nucseq, new_j, best_inserts
        else:
            return cdr3_nucseq, new_j
    ####################################################################################


    if allow_internal_mismatches:
        mismatch_score = -4 # default for junction analysis
    else:
        mismatch_score = -1000
    cdr3_nucseqs = get_degenerate_cdr3_nucseqs(cdr3)

    sortl = []
    for cdr3_nucseq in cdr3_nucseqs:
        result = tcr_sampler.analyze_junction(
            organism, v, j, cdr3, cdr3_nucseq, mismatch_score=mismatch_score,
            return_cdr3_nucseq_src=True)

        inserts = result[-2]
        assert len(inserts) == 4
        total_insert= sum(inserts[1:])
        if verbose:
            print(cdr3, cdr3_nucseq, total_insert, result[0], result[1])
        # ( v_trim, d0_trim, d1_trim, j_trim ) = trims
        # ( best_d_id, n_vd_insert, n_dj_insert, n_vj_insert ) = inserts

        sortl.append((total_insert, cdr3_nucseq, result))
    sortl.sort()

    # now reconstruct top nucleotide sequence
    v_nucseq = tcr_sampler.get_v_cdr3_nucseq( organism, v )
    j_nucseq = tcr_sampler.get_j_cdr3_nucseq( organism, j )
    total_insert, cdr3_degseq, result = sortl[0]

    cdr3_nucseq = list(cdr3_degseq) # can modify in place

    trims, inserts, cdr3_nucseq_src = result[-3:]

    num_matched_v = cdr3_nucseq_src.count('V')
    num_matched_j = cdr3_nucseq_src.count('J')
    num_matched_d = cdr3_nucseq_src.count('D')

    # V
    for i,(a,b) in enumerate(zip(v_nucseq, cdr3_degseq)):
        if i>=num_matched_v:break
        if logo_tools.nucleotide_symbols_match(a,b):
            cdr3_nucseq[i] = a
        else:
            assert allow_internal_mismatches

    # J
    for i,(a,b) in enumerate(zip(reversed(j_nucseq), reversed(cdr3_degseq))):
        if i>=num_matched_j:break
        if logo_tools.nucleotide_symbols_match(a,b):
            cdr3_nucseq[-1-i] = a
        else:
            assert allow_internal_mismatches

    ( v_trim, d0_trim, d1_trim, j_trim ) = trims
    ( best_d_id, n_vd_insert, n_dj_insert, n_vj_insert ) = inserts

    if num_matched_d: # D
        assert best_d_id>0
        assert all_genes[organism][v].chain == 'B'
        d_nucseq = tcr_sampler.all_trbd_nucseq[organism][best_d_id]
        match_start = cdr3_nucseq_src.index('D')
        d_match_seq = d_nucseq[d0_trim:None if d1_trim==0 else -d1_trim]
        assert len(d_match_seq) == num_matched_d
        for i,(a,b) in enumerate(zip(d_match_seq, cdr3_degseq[match_start:])):
            if logo_tools.nucleotide_symbols_match(a,b):
                cdr3_nucseq[match_start+i] = a
            else:
                assert allow_internal_mismatches

    # there may still be degenerate symbols in the N-inserted regions
    for i in range(len(cdr3_nucseq)):
        n = cdr3_nucseq[i]
        if n not in 'acgt':
            cdr3_nucseq[i] = min(logo_tools.nucleotide_classes_lower_case[n],
                                 key=lambda x:N_insertion_frequency_order[x])

    cdr3_nucseq = ''.join(cdr3_nucseq)
    assert all(x in 'acgt' for x in cdr3_nucseq)
    assert translation.get_translation(cdr3_nucseq) == cdr3

    if return_stats:
        return cdr3_nucseq, total_insert
    else:
        return cdr3_nucseq









import numpy as np
import pandas as pd
import multiprocessing as mp

from typing import Union
from functools import lru_cache
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.manifold import MDS

from .repertoire import Repertoire
from .analysis import TcrCollection
from .constants.base import AALPHABET, GAPCHAR
from .constants.hashing import TCRDIST_DM
from .constants.preprocessing import (
    setup_gene_cdr_strings, detect_vgene_col, detect_cdr3_col
)

from sklearn.utils.validation import check_is_fitted

class TCRDistEncoder(BaseEstimator, TransformerMixin):

    def __init__(
        self,
        distance_matrix:np.array=TCRDIST_DM, 
        aa_dim:int=8,
        vgene_col=None,
        cdr3_col=None,
        mds_eps:float=1e-05,
        num_pos:int=16,
        n_trim:int=3,
        c_trim:int=2,
        cdr3_weight:Union[int,float]=3,
        v_weight:Union[int,float]=1,
        organism:str='human',
        chain:str='B',
        full_tcr:bool=True,
        ncpus:int=1
        ):
        """
        TCRDist-based vector embedding for amino acid sequences. Trims and
        gaps sequences to a fixed length and transforms the TCRDist-matrix
        into euclidean space, generating a unique embedding for distinct TCRs
        whose distances reflect the TCRDist-distances between the original sequences.

        Parameters
        ----------
        distance_matrix : np.ndarray[20,20]
            Twenty by twenty matrix containing AA distances. By default, the
            TCRDist matrix is used here.
        aa_dim : int
            Number of dimensions for each amino acid. The final vector will have a
            total length of aa_dim * num_pos dimensions.
        num_pos : int
            Fixed length to which a sequence is gapped after trimming.
        n_trim : int
            Number of amino acids trimmed from left side of sequence.
        c_trim : int
            Number of amino acids trimmed from right side of sequence.
        cdr3_weight : int
            Weighting factor for the CDR3 region relative to other CDRs.
            (Only when including the V gene)
        organism : str
            Organism from which the input sequences originate from.
        chain : str
            TCR chain from which the input sequences originate from.
        full_tcr : bool
            Boolean indicating whether the full TCR (CDR1+CDR2+CDR2.5+CDR3)
            should be used for creating the embedding. This parameter is
            primarily used to calculate the dimensionality of the final vector,
            which is necessary for indexing (see indexing.py classes).
        """
        self.distance_matrix = distance_matrix
        self.aa_dim = aa_dim
        self.vgene_col = vgene_col
        self.cdr3_col = cdr3_col
        self.mds_eps = mds_eps
        self.num_pos = num_pos
        self.n_trim = n_trim
        self.c_trim = c_trim
        self.cdr3_weight = cdr3_weight
        self.v_weight = v_weight
        self.organism = organism
        self.ncpus = ncpus

        allowed_chains = [
            'a','b','g','d',
            'alpha','beta','gamma','delta',
            'ab','gd',
            'alphabeta','gammadelta']
        assert chain.lower() in allowed_chains, f'Invalid chain {chain}, please select alpha, beta, gamma, delta.'
        self.chain = chain
        self.full_tcr = full_tcr

    def __repr__(self):
        return f'TCRDistEncoder(aa_dim={self.aa_dim})'

    def _calc_mds_vecs(self, return_stress=False):
        '''
        Helper function to run MDS.
        '''
        mds = MDS(
            n_components=self.aa_dim,
            eps=self.mds_eps,
            dissimilarity="precomputed",
            random_state=11,
            normalized_stress=False
            )
        vecs = mds.fit_transform(self.dm)
        if return_stress:
            return vecs, mds.stress_
        else:
            return vecs
    
    def _calc_tcrdist_aa_vectors(self, SQRT=True, verbose=False):
        '''
        Embed tcrdist distance matrix to Euclidean space.
        '''
        self.dm = np.zeros((21,21))
        self.dm[:20,:20] = self.distance_matrix
        self.dm[:20,20] = 4.
        self.dm[20,:20] = 4.
        if SQRT:
            self.dm = np.sqrt(self.dm) ## NOTE
        vecs, stress = self._calc_mds_vecs(return_stress=True)
        # print('vecs mean:', np.mean(vecs, axis=0)) #looks like they are already zeroed
        # vecs -= np.mean(vecs, axis=0) # I think this is unnecessary, but for sanity...
        if verbose:
            print(f'encoding tcrdist aa+gap matrix, dim= {self.aa_dim} stress= {stress}')
        return {aa:v for aa,v in zip(AALPHABET+GAPCHAR, vecs)}

    def _trim_and_gap_cdr3(self, cdr3):
        ''' 
        Convert a variable length cdr3 to a fixed-length sequence in a way
        that is consistent with tcrdist scoring, by trimming the ends and
        inserting gaps at a fixed position

        If the cdr3 is longer than num_pos + n_trim + c_trim, some residues will be dropped
        '''
        gappos = min(6, 3+(len(cdr3)-5)//2) - self.n_trim
        r = -self.c_trim if self.c_trim>0 else len(cdr3)
        seq = cdr3[self.n_trim:r]
        afterlen = min(self.num_pos-gappos, len(seq)-gappos)
        numgaps = max(0, self.num_pos-len(seq))
        fullseq = seq[:gappos] + GAPCHAR*numgaps + seq[-afterlen:]
        assert len(fullseq) == self.num_pos
        return fullseq

    # @lru_cache(maxsize=None)
    def _encode_sequence(self, seq):
        '''
        Convert a sequence to a vector by lining up the aa_vectors

        length of the vector will be dim * len(seq), where dim is the dimension of the
        embedding given by aa_vectors
        '''
        # self.calc_tcrdist_aa_vectors()
        dim = self.aa_vectors_['A'].shape[0]
        vec = np.zeros((len(seq)*dim,))
        for i,aa in enumerate(seq):
            vec[i*dim:(i+1)*dim] = self.aa_vectors_[aa]
        return vec

    # @lru_cache(maxsize=None)
    def _gapped_encode_cdr3(self, cdr3):
        '''
        Convert a cdr3 of variable length to a fixed-length vector
        by trimming/gapping and then lining up the aa_vectors

        length of the vector will be dim * num_pos, where dim is the dimension of the
        embedding given by aa_vectors
        '''
        return self._encode_sequence(self._trim_and_gap_cdr3(cdr3))

    def _gapped_encoder_v_cdr3(self, tcr : tuple) -> np.array:
        '''
        Calculate a numerical TCRdist-based encoding for a TCR (V call + junction amino acid)

        tcr, tuple
            Tuple containing V gene and CDR3 amino acid information.
            Example:  ("TRBV19*01", "CASSIGREAFF")
        '''
        v, cdr3 = tcr
        v_vec = np.sqrt(self.v_weight) * self._encode_sequence(self.gene_cdr_strings[v])
        cdr3_vec = np.sqrt(self.cdr3_weight) * self._gapped_encode_cdr3(cdr3)
        return np.concatenate([v_vec, cdr3_vec])
    
    # @lru_cache(maxsize=None)
    def _gapped_encode_tcr_chains(self, tcrs, vgene_col=None, cdr3_col=None) -> np.array:
        '''
        Convert a TCR (V gene + CDR3) of variable length to a fixed-length vector
        by trimming/gapping and then lining up the aa_vectors.

        Parameters
        ----------
        tcrs : pd.DataFrame
            DataFrame with V and CDR3 information in the named columns.
        '''
        # Prepare data structure
        self.tcrs = tcrs
        if vgene_col is None:
            vgene_col = detect_vgene_col(tcrs)
        if cdr3_col is None:
            cdr3_col = detect_cdr3_col(tcrs)
        if isinstance(tcrs, pd.DataFrame):
            tcrs = list(zip(tcrs[vgene_col], tcrs[cdr3_col]))
        else:
            pass
        # !THE FOLLOWING V GENES CONTAIN '*' CHARACTER WHICH IS CAUSING ISSUES WITH THE ENCODING!
        # TRBV12-2*01 -----> FGH-NFFRS-*SIPDGSF
        # TRBV16*02 -------> KGH-S*FQN-ENVLPNSP
        if self.organism == "human":
            to_remove = ['TRBV12-2*01','TRBV16*02']
            n_pre = len(tcrs)
            tcrs = [i for i in tcrs if i[0] not in to_remove]
            n_post = len(tcrs)
            if n_pre-n_post>0:
                print(f"WARNING: Removed TCRs with {to_remove}. This is a temporary measure to prevent KeyError caused by '*' character.\n")
        # Determine the length of the vector
        vec_len = self.aa_dim * (self.num_pos_other_cdrs + self.num_pos)
        # No parallel processing, use a single core
        if self.ncpus == 1:
            vecs = []
            for tcr in tcrs:
                v, cdr3 = tcr
                self._gapped_encoder_v_cdr3(tcr)
                v_vec = np.sqrt(self.v_weight) * self._encode_sequence(self.gene_cdr_strings[v])
                cdr3_vec = np.sqrt(self.cdr3_weight) * self._gapped_encode_cdr3(cdr3)
                vecs.append(np.concatenate([v_vec, cdr3_vec]))
            vecs = np.array(vecs)
        # Parallel processing, use n cores equal to self.ncpus
        else:
            print("mp")
            pool = mp.Pool(processes=self.ncpus)
            vecs = np.array(pool.map(self._gapped_encoder_v_cdr3, tcrs))
            pool.close()
            pool.join()
        # print(len(self.tcrs), vec_len)
        print('Shape of TCR vecs', vecs.shape)
        assert vecs.shape == (len(self.tcrs), vec_len)
        return vecs

    def encode_tcr(self, v, cdr3):
        v_vec = self._encode_sequence(self.gene_cdr_strings[v])
        cdr3_vec = np.sqrt(self.cdr3_weight) * self._gapped_encode_cdr3(cdr3)
        return np.concatenate([v_vec,cdr3_vec])
    
    def _encode_paired_chains(self, tcrs):
        avecs = self._gapped_encode_tcr_chains(tcrs[['va','cdr3a']],'va','cdr3a').astype(np.float32)
        bvecs = self._gapped_encode_tcr_chains(tcrs[['vb','cdr3b']],'vb','cdr3b').astype(np.float32)
        # Concatenate alpha & beta vectors
        abvecs = np.hstack([avecs, bvecs])
        assert abvecs.shape == (tcrs.shape[0], avecs.shape[1] + bvecs.shape[1])
        return abvecs

    def fit(self, X=None, y=None):
        self.aa_vectors_ = self._calc_tcrdist_aa_vectors()
        self.gene_cdr_strings = setup_gene_cdr_strings(self.organism, self.chain)
        self.num_pos_other_cdrs = len(next(iter(self.gene_cdr_strings.values())))
        assert all(len(x)==self.num_pos_other_cdrs for x in self.gene_cdr_strings.values())
        if self.full_tcr:
            self.m = self.aa_dim*self.num_pos + self.aa_dim*self.num_pos_other_cdrs
            if self.chain == 'AB':
                self.m *= 2
        else:
            self.m = self.aa_dim*self.num_pos 
        return self

    def transform(
            self, 
            X: Union[TcrCollection, pd.DataFrame, list, str], 
            split_ab=False,
            vgene_col = 'v_call',
            cdr3_col = 'junction_aa'
            ) -> np.array:
        """
        Generate TCRdist vectors.

        Parameters
        ----------
        x : Union[TcrCollection, list, str]
            Objects to hash, this can be a single CDR3 sequence, but also a
            TcrCollection subclass or list thereof.

        Returns
        -------
        np.array[n,m]
            Array containing m-dimensional hashes for each of the n provided inputs.
        """
        check_is_fitted(self)
        if isinstance(X, (list, np.ndarray)):
            return np.array([self.transform(s) for s in X]).astype(np.float32)
        elif isinstance(X, pd.DataFrame):
            if self.full_tcr:
                if self.chain == 'AB':
                    # if not in paired format
                    if not set(['va','vb','cdr3a','cdr3b']).issubset(X.columns):
                        assert 'locus' in X.columns, f"DataFrame must include column named 'locus'."
                        rep = Repertoire(X)
                        X = rep.airr_to_tcrdist_paired()
                    # split up alpha and beta vecs
                    if split_ab:
                        avecs = self._gapped_encode_tcr_chains(X, 'va', 'cdr3a').astype(np.float32)
                        bvecs = self._gapped_encode_tcr_chains(X, 'vb', 'cdr3b').astype(np.float32)
                        return avecs, bvecs
                    else:
                        return self._encode_paired_chains(X)
                else:            
                    # assert 'v_call' in X.columns, f"DataFrame is missing 'v_call' column."
                    # assert 'junction_aa' in X.columns, f"DataFrame is missing 'junction_aa' column."
                    return self._gapped_encode_tcr_chains(X,self.vgene_col,self.cdr3_col).astype(np.float32)
            else:
                assert 'junction_aa' in X.columns, f"DataFrame does not include column named 'junction_aa'."
                X = X.junction_aa.to_list()
                return np.array([self.transform(s) for s in X]).astype(np.float32)
        else:
            return self._gapped_encode_cdr3(X).astype(np.float32)
        
def join_ab_vecs(avecs,bvecs):
    return np.hstack([avecs, bvecs])


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from . import _version
__version__ = _version.get_versions()['version']

from .datasets import load_test
from .neighbors import neighbor_analysis



import pandas as pd

from os.path import join, dirname, abspath

DIR = dirname(abspath(__file__))

def load_test(column_type='single'):
    if column_type == 'single':
        return pd.read_table(join(DIR, 'data/test_single.tsv'))
    elif column_type == 'paired':
        return pd.read_table(join(DIR, 'data/test_paired.tsv'))



#from . import amino_acids
from .amino_acids import amino_acids
import sys

def get_alphabet( pwm ):
    alphabet = list(pwm[0].keys())[:]
    alphabet.sort()
    return alphabet

def check_pwm( pwm, tol= 0.001 ):
    L = len(pwm)
    alphabet = get_alphabet( pwm )
    for pos in range(L):
        for aa,val in pwm[pos].items(): assert val > -1e-6
        total = sum( ( pwm[pos][aa] for aa in alphabet ) )
        assert abs( total - 1.0 ) - tol

def create_protein_pwm_from_sequences( seqs, pseudocounts = 0.0 ):
    return create_pwm_from_sequences( seqs, amino_acids, pseudocounts=pseudocounts )

def create_dna_pwm_from_sequences( seqs, pseudocounts=0.0 ):
    return create_pwm_from_sequences( seqs, list('acgt'), pseudocounts=pseudocounts )

def create_pwm_from_sequences( seqs, alphabet, pseudocounts=0.0 ):
    pwm = {}
    if len(seqs) == 0: return pwm
    L = len( seqs[0] )

    for pos in range(L):
        pwm[ pos ] = dict( list(zip( alphabet, [pseudocounts]*len(alphabet) )) )

    for s in seqs:
        assert len(s) == L
        for pos in range(L):
            if s[pos] not in alphabet:
                sys.stderr.write('logo_tools.create_pwm_from_sequences: skipping bad character %s\n'%(s[pos]))
                continue
            pwm[ pos ][ s[pos] ] += 1

    for pos in range(L):
        norm = 1.0 / sum( pwm[pos].values() )
        for a in alphabet: pwm[ pos ][ a ] *= norm

    check_pwm( pwm )

    return pwm



base_partner = {'a': 't', 'c': 'g', 'g': 'c', 't': 'a', 'n': 'n',
                'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N',
                'R': 'Y', 'Y': 'R',
                'S': 'S', 'W': 'W',
                'K': 'M', 'M': 'K',
                '.': '.' }

nucleotide_classes_lower_case = { 'a':'a',
                                  'c':'c',
                                  'g':'g',
                                  't':'t',
                                  'w':'at',
                                  's':'cg',
                                  'k':'gt',
                                  'm':'ac',
                                  'y':'ct',
                                  'r':'ag',
                                  'b':'cgt',
                                  'd':'agt',
                                  'h':'act',
                                  'v':'acg',
                                  'n':'acgt' }

nuc_match_lower_case = {}

for x in nucleotide_classes_lower_case:
    for y in nucleotide_classes_lower_case:
        nuc_match_lower_case[(x,y)] = False
        for x1 in nucleotide_classes_lower_case[x]:
            for y1 in nucleotide_classes_lower_case[y]:
                if x1==y1:
                    nuc_match_lower_case[(x,y)] = True
                    break

def nucleotide_symbols_match( a_in, b_in ):
    a = a_in.lower()
    b = b_in.lower()
    if a==b: return True
    return nuc_match_lower_case.get( (a,b), False )

def degnucs_contain(a,b):
    ''' Check if a contains b, allowing for degenerate nucleotide symbols
    '''
    #
    if len(b)>len(a):
        return False
    for offset in range(len(a)-len(b)):
        if all(nucleotide_symbols_match(x,y)
               for x,y in zip(a[offset:], b)):
            return True
    return False

def degnucs_index(a,b):
    ''' Where does b match within a?
    assert failure if not degnucs_contain(a,b)
    '''
    #
    assert len(b)<=len(a), 'degnucs_index: b is not contained in a'
    for offset in range(len(a)-len(b)):
        if all(nucleotide_symbols_match(x,y)
               for x,y in zip(a[offset:], b)):
            return offset
    assert False, 'degnucs_index: b is not contained in a'
    return 0

def reverse_complement( seq ):
    newseq = ''
    L = len(seq)
    for pos in range( L-1, -1, -1 ):
        newseq += base_partner[ seq[ pos ] ]
    assert len( newseq ) == L
    return newseq

